// partial backbone
var ExecuteButtonView = Backbone.View.extend({
    el: $("#execute"),
    initialize: function(){
        this.model.on("change", this.switchFilename, this);
        this.model.on("change", this.switchButtonText, this);
        this.switchFilename();
        this.switchButtonText();
    },
    events: {
        "click": "onclick"
    },
    switchButtonText: function(){
        this.$el.text(app.get("isRunning") ? "Stop" : "Execute");
    },
    onclick: function(event){
        if (this.model.get("currentFileName") === ""){
            app.set("execstat", "Save the network in advance");
            return;
        }

        var that = this;

        if (app.get("isRunning")){
            // kill the process
            app.set("isRunning", false);
            $.ajax({
                url: rootPath + "abort"
            }).done(function(data){
                app.set("execstat", data);
            }).fail(function(data, status, error){
                console.log(data, status, error);
                app.set("execstat", "Error " + data.responseText);
            });
        } else {
            // run the process
            var executor = function(filename){
                app.set("isRunning", true);
                $.ajax({
                    url: "/preference/environment",
                    method: "PUT",
                    data: userPreference.get("env")})
                .always(function(data){
                    $.ajax({
                        url: rootPath + "execution/" + app.get("currentFileName"),
                        cache: false
                    }).done(function(data){
                        console.log(data);
                        app.set("execstat", data);
                    }).fail(function(data, status, error){
                        console.log(data, status, error);
                        app.set("execstat", "Error " + data.responseText);
                    });
                    setTimeout(polling, 1000);
                });
            };

            if (app.get("currentFileName") === ""){
                $("#saveFileDialog").dialog("dialogclose", executor);
                $("#saveFileDialog").dialog("open");
            } else {
                $.ajax({
                    url: rootPath + "file/object/" + app.get("currentFileName"),
                    type: "PUT",
                    data: network,
                    // dataType: "json"
                }).done(function(){
                    app.set("execstat", "Saved");
                    updateFileList();
                }).fail(function(data){
                    console.log("failed", data);
                    app.set("execstat", data.responseText);
                });
                executor();
            }
        }
    },
    switchFilename: function(){
        if (this.model.get("currentFileName") !== ""){
            this.$el.addClass("menuElement");
            this.$el.removeClass("menuElementDisabled");
        } else {
            this.$el.addClass("menuElementDisabled");
            this.$el.removeClass("menuElement");
        }
    }
});
function polling(){
    console.log("polling");
    var that = this;
    $.ajax({
        url: rootPath + "execstatus"
    }).done(function(data){
        app.set("execstat", data);
        if ("Process is running." === data){
            console.log("register again");
            setTimeout(that.polling, 1000);
        } else {
            app.set("isRunning", false);
        }
    });
}

var ZoomView = Backbone.View.extend({
    el: $("#zoomSlider"),
    initialize: function(){
        this.listenTo(this.model, "change", this.update);
        this.render();
    },
    render: function(){
        var that = this;
        that.$el.slider({
            value: that.model.get("zoom"),
            min: 1,
            max: 10,
            slide: function(event, ui){
                // replace class & calculate the size ratio
                var baseNode = $(".node:visible:first");
                var prev = baseNode.width();
                console.log(that.model.get("zoom"));
                $(".node").removeClass("nodeSize" + that.model.get("zoom"));
                $(".node").addClass("nodeSize" + ui.value);
                that.model.set("zoom", ui.value);
                var aft = baseNode.width();
                var ratio = 1.0 * aft / prev;

                // update node.pos
                // Note: leftest and toppest is kept same
                _.each(_.groupBy(network.nodes, "sheet"), function(nodes){
                    var sheetname = nodes[0].sheet;

                    // leftest/toppest before moving
                    var offsets = _.map(nodes, function(n){return {top: n.pos[1], left: n.pos[0]};});
                    var leftestPrev = _.min(_.pluck(offsets, "left"));
                    var toppestPrev = _.min(_.pluck(offsets, "top"));

                    _.each(nodes, function(n){
                        n.pos[0] *= ratio;
                        n.pos[1] *= ratio;
                    });

                    // leftest/toppest after moving
                    offsets = _.map(nodes, function(n){return {top: n.pos[1], left: n.pos[0]};});
                    var leftestAft = _.min(_.pluck(offsets, "left"));
                    var toppestAft = _.min(_.pluck(offsets, "top"));

                    // shift nodes
                    _.each(nodes, function(n){
                        n.pos[0] = truncate(n.pos[0] + leftestPrev - leftestAft);
                        n.pos[1] = truncate(n.pos[1] + toppestPrev - toppestAft);
                    });

                    // update offset if the sheetname is the currentTab
                    if (sheetname === currentTab){
                        _.each(nodes, function(n){$("#" + n.id).offset({left: n.pos[0], top: n.pos[1]});});
                        clearCanvas();
                        drawLinks();
                    }
                });
            }
        });
    },
    update: function(){
        this.$el.slider("value", this.model.get("zoom"));
    }
});

var UserPreferenceModel = Backbone.Model.extend({
    defaults: {
        size: "normal",
        packages: [],
        nodelistmin: false,
        colors  : [],
        link    : "Bezier",
        zoom    : 4,
        lang    : "en",
        showTooltip: "Yes",
        consoleLeft: "",
        consoleTop : "",
        env : ""
    },
    initialize: function(){
        this.on("change", this.save);
        this.on("change", this.updateDOMs);
        if (localStorage.userPref){
            var value = JSON.parse(localStorage.userPref);
            var attrs = ["showTooltip",
                         "size",
                         "colors",
                         "link",
                         "zoom",
                         "lang",
                         "consoleLeft",
                         "consoleTop",
                         "env"];
            for (var i = 0; i < attrs.length; i++){
                this.set(attrs[i], value[attrs[i]]);
            }
            var pipes = [];
            for (var i = 0; i < value.packages.length; i++){
                this.addPackage(value.packages[i]);
            }

            $("#console").css("left", value.consoleLeft);
            $("#console").css("top",  value.consoleTop );
            $(document).tooltip();
            if (this.get("showTooltip") === "No"){
                $(document).tooltip("disable");
            }
        } else {
            // if no localStorage found, load all packages
            $.ajax({
                url: rootPath + "packagelist",
                method: "GET"
            }).done(function(data){
                _.each(data, function(p){
                    userPreference.addPackage(p);
                });
            });
        }

        this.setSelectValues();
        var that = this;
        $("#preferenceDialog .display button").button().click(function(){
            that.set("lang", "en");
            that.set("link", "Bezier");
            that.set("showTooltip", "Yes");
            that.setSelectValues();
            that.trigger("change");
        });
        var maxwidth = _.chain($("#preferenceDialog .display label"))
                        .map(function(l){return $(l).width();})
                        .max().value();
        _.each($("#preferenceDialog .display label"), function(l){
            $(l).css("width", (maxwidth + 20) + "px");
            $(l).css("display", "inline-block");
        });
    },
    setSelectValues: function(){
        $("#preferenceDialog .display select[name=lang]").val(this.get("lang"));
        $("#preferenceDialog .display select[name=line]").val(this.get("link"));
        $("#preferenceDialog .display select[name=tooltip]").val(this.get("showTooltip"));
    },
    updateDOMs: function(){
        if (typeof(userPreference) !== "undefined"){
            setupIntroJS();
        }
    },
    save: function(){
        localStorage.userPref = JSON.stringify(this.toJSON());
    },
    addPackage: function(def){
        var p = this.get("packages");
        if (_.indexOf(p, def) === -1){
            this.set("packages", p);
            loadNodeListAjax(def);
            p.push(def);
            this.save();
        }
        nodelistSize();
    },
    delPackage: function(def){
        userPreference.set("packages",
            _.without(userPreference.get("packages"), def));
        nodedefs = _.reject(nodedefs, function(nodedef){
            return nodedef.def === def;
        });
        this.save();
        nodelistSize();
    }
});

var NowLoading = Backbone.View.extend({
    el: $(".nodeLoading"),
    initialize: function(){
        this.listenTo(this.model, "change", this.render);
    },
    render: function(){
        // if (_.isNumber(this.model.get("loadStage"))){
        //     this.$el.html("Now loading... " + this.model.get("loadStage") + "%");
        // }
        // else {
        //     this.$el.html("Now loading... " + this.model.get("loadStage"));
        // }
    }
});

var App = Backbone.Model.extend({
    defaults: {
        isRunning: false,
        execstat: "",
        loadStage: 0,
        currentFileName: "",
        clientOS: "",
        serverOS: ""
    }
});

var ConsoleView = Backbone.View.extend({
    el: $("#console"),
    prev: "",
    initialize: function(){
        this.listenTo(this.model, "change", this.render);
        this.$el.draggable({
            cancel: "#executeStatus",
            containment: "body",
            drag: function(){
                userPreference.set("consoleLeft", $(this).css("left"));
                userPreference.set("consoleTop",  $(this).css("top"));
            }
        });
        this.$el.zIndex(10);
    },
    render: function(){
        var newtext = this.model.get("execstat")
                                .replace(/<stderr>/g, "<span style='color:#900'>")
                                .replace(/<\/stderr>/g, "</span>")
                                .replace(/<stdout>/g, "<span style='color:#000'>")
                                .replace(/<\/stdout>/g, "</span>")
                                .replace(/\n/g, "<br/>")
                                .replace(/<br\/><br\/>/g, "<br/>");
        if (newtext.length > 100){
            var tmp = newtext.split("<br/>")[0];
            this.$el.find("#executeStatus").html(tmp + "<span style='margin-left:10px;'>see more</span>");
            var seemore = $(this.$el.find("#executeStatus").find("span"));

            seemore.css("text-decoration", "underline");
            seemore.attr("onclick",
                'window.open("data:text/html;charset=utf-8,' +
                encodeURIComponent(newtext) + '", "_blank");');
            tmp += "<span style=''" +
                   " onclick=\'window.open('data:text/html;charset=utf-8'" +
                    encodeURI(newtext) + "')\'>see more</span>";
            newtext = tmp;
        } else {
            this.$el.find("#executeStatus").html(newtext);
        }

        if (this.prev !== this.model.get("execstat")){
            if (this.model.get("execstat").indexOf("Error") !== -1){
                this.$el.switchClass("consoleNoerror", "consoleError", 400, "linear")
                        .delay(1000)
                        .switchClass("consoleError", "consoleNoerror", 400, "linear");
            } else {
                this.$el.switchClass("consoleError", "consoleNoerror", 400, "linear");
            }
        }
        $("#saveFileDialog input[name=name]").val(this.model.get("currentFileName"));
        $("#currentFileName").html(this.model.get("currentFileName"));
        this.prev = this.model.get("execstat");
    }
});

var FileModel = Backbone.Model.extend({
    defaults: {
        type: "",
        name: "",
        date: new Date(),
        object: ""
    }
});

var FileCollection = Backbone.Collection.extend({
    model: FileModel
});

var FileView = Backbone.View.extend({
    tagName: "tr",
    className: function(){
        return this.model.get("type");
    },
    events: {
        "mouseover": "onMouseover",
        "mouseout" : "onMouseout",
        "mousedown": function(e){return false;}
    },
    onMouseover: function(e){
        // this.$el.addClass("ui-widget");
        // this.$el.addClass("ui-state-default");
        // this.$el.addClass("ui-corner-all");
    },
    onMouseout: function(e){
        // this.$el.removeClass("ui-widget");
        // this.$el.removeClass("ui-state-default");
        // this.$el.removeClass("ui-corner-all");
    },
    initialize: function(){
        this.listenTo(this.model, "remove", this.remove);
    },
    render: function(){
        var str = "";
        str += "<td><button class='delete' name='<%= name %>'>Delete</button></td>" +
               "<td class='type'><%= type %></td>" +
               "<td class='name'><%= name %></td>" +
               "<td class='buttonset'>" +
               "    <a class='download' href='dl/network/<%= name %>' target='_blank'>Download</a>";
        if (this.model.get("type") === "network"){
            str += "    <button class='load' name='<%= name %>'>Load</button>";
        }
        str += "</td>";

        this.$el.append(_.template(str,this.model.toJSON()));
        // console.log(_.template(str,this.model.toJSON()));

        this.$el.find(".download").button({
            text: false,
            icons: {primary: "ui-icon-arrowthickstop-1-s"}
        });
        this.$el.find(".load").button({
            text: false,
            icons: {primary: "ui-icon-folder-open"}
        }).click(function(){
            if (confirm("Current network will be cleaned. OK?")){
                loadFile($(this).attr("name"));
            }
        });
        this.$el.find(".delete").button({
            text: false,
            icons: {primary: "ui-icon-closethick"}
        }).click(function(){
            if (confirm("Are you sure to delete file" + $(this).attr("name") + "?")){
                deleteFile($(this).attr("name"));
            }
        });

        this.$el.find(".load").tooltip({position: {my: 'left center', at: 'right+5 center'}});
        this.$el.find(".download").tooltip({position: {my: 'left center', at: 'right+5 center'}});
        this.$el.find(".delete").tooltip({position: {my: 'right center', at: 'left-5 center'}});

        this.$el.css("list-style-type", "none");
        return this;
    }

});
var debug;
var FileManagerView = Backbone.View.extend({
    el: $("#fileManagerDialog"),
    initialize: function(){
        var that = this;
        var table = that.$el.children(".dialogMain");
        $("#fileProgress").progressbar();
        this.$el.find(".reload").button({
            text: false,
            icons: {primary: "ui-icon-arrowrefresh-1-s"}
        }).click(function(e){updateFileList();});

        $('#fileUploadFile').fileupload({
            // dataType: "json",
            url: function(){ return $("#fileUploadFile").attr("data-url");},
            done: function(e, data){
                updateFileList();
                data.context.unbind().remove();
            },
            progressall: function(e, data){
                var progress = parseInt(data.loaded / data.total * 100, 10);
                if (progress === 100){
                    $('#fileProgress').css("display", "none");
                } else {
                    $('#fileProgress').css("display", "block");
                }
                $("#fileProgress").progressbar("value", progress);
            },
            add: function(e, data){
                if (_.pluck(fileCollection.toJSON(), "name").indexOf(data.files[0].name) !== -1){
                    if (!confirm("Overwrite " + data.files[0].name + "?")){
                        $("#fileManagerLoading").css("display", "none");
                        return;
                    }
                }
                data.context =
                    $(_.template($("#fileUploadTemplate").html(), data).trim())
                    .appendTo("#uploadList");
                function checkType(filename){
                    var ext = filename.split(".").slice( -1 )[0];
                    if (ext === "n"){
                        return "network";
                    }
                    if (_.indexOf(["dat", "tff"], ext) !== -1){
                        return "tf";
                    }
                    if (_.indexOf(["wav", "raw"], ext) !== -1){
                        return "audio";
                    }
                    return "others";
                }
                var kind = checkType(data.files[0].name);

                data.url = "/file/" + kind + "/" + data.files[0].name + "/" + data.files[0].size;
                data.submit();
                data.context.unbind().remove();
            },
            dropZone: $("#fileManagerDialog")
        });

        // add data path
        $.ajax({
            url: rootPath + "datapath",
            method: "GET"
        }).done(function(res){
            $("#datapath").val(res);
            $("#datapath").on("click", function(){this.select(0, this.value.length);});
        });

        // setup dialog
        this.$el.dialog({ autoOpen: false,
                          modal: true,
                          minWidth: 710 });

        $("#fileManagerDialog .dialogMain").on("mousedown", function(){return false;});

        // this.listenTo(this.collection, "all", function(e, b){console.log(e, b.toJSON());});
        this.listenTo(this.collection, "add", this.addFile);
        this.listenTo(this.collection, "reset", this.reset);

        this.listenTo(this.collection, "add", this.update);
        this.listenTo(this.collection, "reset", this.update);
        this.listenTo(this.collection, "delete", this.update);
        this.on("close", this.close);
    },
    arrangeSize: function(type){
        _.each($("#file_" + type), function(d){
            var filenames = $(d).find("span.name");
            if (filenames.length === 0){return;}

            var width = _.chain(filenames)
                         .map(function(s){return $(s).width();})
                         .max().value();
            if (width !== -Infinity){
                _.each(filenames, function(f){$(f).css("width", width + "px");});
            }
        });
    },
    reset: function(resetted, m){
        _.each(m.previousModels, function(m){
            m.trigger("remove");
            m.destroy();
        });
    },
    close: function(){
        this.$el.dialog("close");
    },
    addFile: function(file){
        var view = (new FileView({model: file})).render();
        $(this.$el.find(".dialogMain")).append(view.el);
    },
    update: function(file){
        var that = this;
        var exists = _.chain(this.collection.toJSON()).groupBy("type").keys().value();
        _.each(this.$el.find(".ui-tabs-anchor"), function(a, i){
            if (_.indexOf(exists, $(a).attr("type")) === -1){
                $("#file_" + $(a).attr("type")).children(".nofile").text("no files");
            } else {
                $("#file_" + $(a).attr("type")).children(".nofile").text("");
            }
        });
    }
});
var debug;
var SheetsView = Backbone.View.extend({
    el: $("#sheets"),
    initialize: function(){
        this.listenTo(this.model, "change", this.toggleMinimize);
        this.toggleMinimize();
    },
    toggleMinimize: function(event, prev){
        if (this.model.get("nodelistmin")){
            this.$el.css("margin-left", "0px");
        } else {
            $("#sidebar").css("width", "auto");
            $("#sheets").css("margin-left", $("#sidebar").width() + "px");
            $("#sidebar").css("width", $("#sidebar").width() + "px");
            // this.$el.css("margin-left", "200px");
        }
    }
});

var NodelistView = Backbone.View.extend({
    el: $("#sidebar"),
    initialize: function(){
        this.listenTo(this.model, "change", this.toggleMinimize);
        this.toggleMinimize();
    },
    toggleMinimize: function(event){
        if (this.model.get("nodelistmin")){
            this.$el.css("display", "none");
        } else {
            this.$el.css("display", "block");
        }
    }

});
var MinimizeNodelistButton = Backbone.View.extend({
    el: $("#minimizeNodelistButton"),
    events: {
        "click": "onclick"
    },
    initialize: function(){
        this.listenTo(this.model, "change", this.switchIcon);
        this.switchIcon();
    },
    onclick: function(event){
        this.model.set("nodelistmin", !this.model.get("nodelistmin"));
    },
    switchIcon: function(event){
        if (this.model.get("nodelistmin") === false){
            this.$el.button({label: "Hide list", text: false});
        } else {
            this.$el.button({label: "Show list", text: false});
        }
    }

});
var zoomView;
var executeButtonView;
var fileManagerView;
var consoleView;
var nowLoadingView;
var sheetsView;
var nodelistView;
var minimizeNodelistButton;

function setupAllBackbone() {
    app = new App();
    userPreference = new UserPreferenceModel();
    fileCollection     = new FileCollection();

    zoomView = new ZoomView({model: userPreference});
    executeButtonView = new ExecuteButtonView({model: app});
    fileManagerView = new FileManagerView({collection: fileCollection});
    consoleView = new ConsoleView({model: app});
    nowLoadingView = new NowLoading({model: app});
    sheetsView = new SheetsView({model: userPreference});
    nodelistView = new NodelistView({model: userPreference});
    minimizeNodelistButton = new MinimizeNodelistButton({model: userPreference});
}

/* Global variables */
var prevQuery = "";
var network = newNetwork();
var nodedefs = [];
var selectedNodeObj;
var currentTab = "MAIN";
var currentLinkFrom = false;
var clipboard;
var isDraggingStarted = false;
var scrolls = {};
var isSearching = false;

var rootPath = $("#rootPath").attr("root") || "/";
var isSelenium = $("#rootPath").attr("selenium");

// delimiter related definitions
// used to connect node_id and terminal name. e.g., node_Synthesize_1[delim]OUTPUT
delim = "_delim_";
function getBeforeDelim(str){return str.substring(0, str.lastIndexOf(delim));}
function getAfterDelim(str) {return str.substring(str.lastIndexOf(delim) + delim.length);}

$(document).ready(function() {
    // is logged in?
    // if (localStorage.loggedin !== "true"){
    //     window.onbeforeunload = undefined;
    //     alert("you are not logged in");
    //     window.location = "http://localhost:3000/login";
    //     return;
    // }

    if (isSelenium === "right"){
        window.onbeforeunload = function(event){
            event = event || window.event;
            if (event){
                event.returnValue = "Do you really want to change the page?";
            }
            return "Do you really want to change the page?";
        };
    }

    // jQuery UI widgets setup
    $("#sidebar").accordion({
        active: 1,
        icons: false
    });
    $("#nodelist").menu();
    $("#flatNodelist").menu();
    $("#filelist").menu();
    $(document).on("focus", "#nodelistSearch input", function(){
        $(".nodeSelected").removeClass("nodeSelected");
        selectedNodeObj = undefined;
    });

    setupAllBackbone();
    $.ajax({
        url: rootPath + "os"
    }).done(function(data){
        app.set("serverOS", data);
    });
    setupSheets();
    setupNodeEvents();
    setupNodeKeyBindings();
    setupSaveFile();
    setupTerminalRoles();
    setupParameterValidation();
    setupButtons();
    setupStatus();
    setupPreference();
    setupFileManager();
    setupIntroJS();
    if (userPreference.get("env")){
        $.ajax({
            url: "/preference/environment",
            method: "PUT",
            data: userPreference.get("env")});
    }
    if ($("#logout").length){
       $("#logout").button().click(function(){
           $.ajax({
                url: "/logout"
            }).always(function(data){
                console.log("logout", data);
                window.location = "/login";
            });
        });
    }

    addSheetScrollEvent("MAIN");
    _.delay(clearCanvas, 400);
    $(".loading").delay(200).fadeOut("slow")
                 .queue(function(){
                    // $(".loading").remove();
                    $("#nodelistSearch input").focus();
                    nowLoadingView.unbind().remove();
                 });
    keepalive();
});
var interval = 5000;
function keepalive(){
    var msg = "Server connection failed";
    $.ajax({
        url: rootPath + "keepalive",
        cache: false
    }).done(function(data){
            interval = 5000;
            if (app.get("execstat") === msg){
                app.set("execstat", "Server connection established");
            }
            setTimeout(keepalive, interval);
    }).fail(function(data){
            interval = _.min([interval * 2, 60000]); // max 1min
            app.set("execstat", msg);
            setTimeout(keepalive, interval);
    });
}
var timer = false;
$(window).resize(function(){
    // console.log("resizeing");
    // if (timer !== false){
    //     clearTimeout(timer);
    // }
    // timer = setTimeout(function(){
        updateSheetsSize();
        clearCanvas();
        drawLinks();
    // }, 200);
});

function setupTerminalRoles(){
    $("#terminalRoleNameDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            OK: function(){
                // var clickedTermId = $(this).attr("clickedTermId");
                // _.find(network.terminalRoles, function(role){
                //     return clickedTermId ===
                //            role.nodeId + delim + role.termName;
                // }).name = $("#terminalRoleName").val();
                newTerminalRole({
                    nodeId  : $(this).attr("clickedNodeId"),
                    termName: $(this).attr("clickedTermName"),
                    name    : $("#terminalRoleName").val(),
                    role    : $(this).attr("role"),
                    sheet   : currentTab,
                    object_type: $(this).attr("ioobj_type"),
                    description: $(this).attr("ioobj_description")
                });
                $(this).dialog("close");
            },
            Cancel: function(){
                $("#terminalRoleName").val("");
                $(this).dialog("close");
            }
        }
    });
    $("#terminalRoleNameDialog").find("form").submit(function(event){
        $($('#terminalRoleNameDialog').siblings('.ui-dialog-buttonpane').find("button")[0]).click();
        event.preventDefault();
        event.stopPropagation();
    });
}

/* netwrok data structure construction */
function reset(){
    currentTab = "MAIN";
    currentLinkFrom = false;
    app.set("currentFileName", "");

    // remove all Dynamic nodedefs
    nodedefs = _.reject(nodedefs, function(def){
        return def.category === "Dynamic";
    });
    app.set("execstat", "");

    // remove all sheets except MAIN
    $("#sheets").tabs({active: 0});
    _.each(_.keys(network.sheets),
        function(sheetname){
            if (sheetname !== "MAIN"){
                removeSheet("sheet_" + sheetname, true);
            }
    });

    // remove all nodes in MAIN
    _.each(network.nodes, function(n){removeNode(n.id, false, true);});
    clearCanvas("canvas_MAIN");
    addSheetScrollEvent("MAIN");
}

function newNetwork(){
    return {"nodes": [],
            "links": [],
            "terminalRoles": [],
            "sheets": {"MAIN": {name: "MAIN", type: "subnet"}}
            };
}

function nodeId2nodename(nodeId){
    var pre = "node_";
    var post = "_";
    return nodeId.substring(nodeId.indexOf(pre) + pre.length,
                            nodeId.lastIndexOf(post));
}

function loadFile(filename){
    $("#fileManagerLoading").css("display", "inline");
    $.ajax({
        url: rootPath + "file/" + filename,
        method: "GET",
        dataType: "JSON"
    }).done(function(data){
        if (data.type !== "network"){
            app.set("execstat", "Data type is not network");
            return;
        }
        reset();
        app.set("execstat", "Clearing finished. loading...");
        _parseData(data);
        app.set("execstat", "Loading finished.");
    }).fail(function (data){
        updateFileList();
        app.set("execstat", "Loading file failed.");
    }).always(function(data){
        $("#fileManagerLoading").css("display", "none");
    });
    function _parseData(data){
        network = data.object;

        pruneNetwork(network);
        app.set("currentFileName", data.name);

        // add sheet
        _.each(network.sheets, function(sheet){
            if (sheet.name !== "MAIN"){
                addSheetDOM(sheet);
                addDynamicNode(sheet.name);
            }
        });

        // network pruning
        network.nodes = _.map(network.nodes, function(n){
            var def = _.findObject(nodedefs, "name", n.name) || {inputs: [], outputs: []};
            if (typeof(def.name) === "undefined"){
                console.log("undefined node found:" + JSON.stringify(def));
            }
            n.def = def.def || "Dynamic";
            n.theme = def.theme || "Classic";
            n.category = def.category || "Undefined";

            // prune inputs/outputs
            n.inputs  = _.uniq(_.union(def.inputs,  n.inputs  || []), function(io){return io.name;});
            n.outputs = _.uniq(_.union(def.outputs, n.outputs || []), function(io){return io.name;});

            // offset;
            n.pos[0] += $("#sheets").offset().left + 100;
            n.pos[1] += $("#sheets").offset().top  + 100;

            // update parameters if exists in the network.
            if (n.def === "Dynamic"){
                n.parameters = n.parameters || [];
            } else if (def.parameters){
                n.parameters = _.map(def.parameters, function(dp){
                    var param = _.find(n.parameters, function(np){return np.name === dp.name;});
                    if (param) {
                        param = _.extend(dp, {value: param.value, type: param.type});
                    } else {
                        param = dp;
                    }
                    return param;
                });
                n.parameters = JSON.parse(JSON.stringify(n.parameters));
            }
            // prune parameters
            _.each(n.parameters, function(p){
                p.condition = p.condition || "";
                p.list = p.list || [];
            });

             return n;
        });

        _.each(network.links, function(l){
            var addLinkedIOs = function (nodeId, termName, io){
                _.chain(network.nodes)
                 .filter(function(n){return n.id === nodeId;})
                 .each(function(n){
                    // if links connects nonexistent term, add it to the node
                    if (_.pluck(n[io], "name").indexOf(termName) === -1){
                        n[io].push({
                            description: io + "added by the user",
                            name: termName,
                            type: "any"
                        });
                    }
                 });
             };
            addLinkedIOs(l.fromNodeId, l.fromTermName, "outputs");
            addLinkedIOs(l.toNodeId,   l.toTermName,   "inputs");
        });

        // terminalRoles to output
        _.each(network.terminalRoles, function(r){
            var addTerminalRolesIO = function (nodeId, termName, io){
                _.chain(network.nodes)
                 .filter(function(n){return n.id === nodeId;})
                 .each(function(n){
                    // if links connects nonexistent term, add it to the node
                    if (_.pluck(n[io], "name").indexOf(termName) === -1){
                        console.log(io, n[io]);
                        n[io].push({
                            description: io + "added by the user",
                            name: termName,
                            type: "any"
                        });
                    }
                });
            };
            if (r.role === "input" || r.role === "output"){
                addTerminalRolesIO(r.nodeId, r.termName, r.role + "s");
            }
        });

        // dom
        _.each(network.nodes, function(n){
            addNodeDOM(n);
        });

        // add terminalRoles for non-MAIN sheet
        updateDynamicNodeDOM(network);
        _.chain(network.terminalRoles)
         .reject(function(r){return r.sheet === "MAIN";})
         .each(function(r){
            updateTerminalRoles(r, "add");
        });

        // add terminalRoles for MAIN sheet
        _.chain(network.terminalRoles)
         .filter(function(r){return r.sheet === "MAIN";})
         .each(function(r){
            updateTerminalRoles(r, "add", true);
         });

        _.each(_.keys(network.sheets), function(sheet){
            clearCanvas("canvas_" + sheet);
            drawLinks("canvas_" + sheet);
        });

        updateDynamicNodeDOM(network);

        $("#fileManagerLoading").css("display", "none");
        _.chain(network.sheets).keys().each(function(s){adjustScroll(s);});

        //DOM position
        console.log("DOM position update");
        _.chain(network.node).filter(function(n){return n.sheet === currentTab;})
         .each(function(n){$("#" + n.id).offset({left: n.pos[0], top: n.pos[1]});});
        fileManagerView.trigger("close");
    }
}

function deleteFile(filename){
    $.ajax({
        url: rootPath + "file/" + filename,
        method: "DELETE"
    }).done(function(data){
        app.set("execstat", "Deleted");
        console.log(data);
        updateFileList();
    });
}

/* setup nodes events */
function addNode(nodename, sheetname){
    // memo: nodecount starts from 1.
    var samenames = _.filter(network.nodes, function(n){
        return n.name === nodename;
    });
    console.log(samenames);

    var nodecount = 1;
    if (samenames.length > 0){
        var maxindex = _.chain(samenames).pluck("index").max().value() + 2;
        console.log(maxindex);
        nodecount = _.find(_.range(1, maxindex), function(candidate){
            console.log(_.pluck(samenames, "index"));
            console.log(samenames);
            console.log(candidate);
            if (_.indexOf(_.pluck(samenames, "index"), candidate) === -1){
                return true;
            }
        });
        console.log(nodecount);
    }
    // _.each(network.nodes, function(node){
    //     return node.name == nodename;
    // }).length + 1;
    var nodeId = "node_" + nodename + "_" + nodecount;
    var nodedef = _.findObject(nodedefs, "name", nodename);

    // add to the data structure
    newNode({
        name: nodename,
        id: nodeId,
        sheet: sheetname,
        index: nodecount,
        pos: undefined,
        parameters: JSON.parse(JSON.stringify(nodedef.parameters)), // copy
        inputs: JSON.parse(JSON.stringify(nodedef.inputs)),
        outputs: JSON.parse(JSON.stringify(nodedef.outputs)),
        def: nodedef.def,
        theme: nodedef.theme,
        category: nodedef.category
    });
    var addedNode = _.last(network.nodes);

    addNodeDOM(addedNode);
}

function addNodeDOM(nodeobj, noAnimation){
    var nodeId   = nodeobj.id;
    var nodename = nodeobj.name;
    var nodedom, nodeparamdom;
    var sheetname  = nodeobj.sheet;
    var existingTerminalRoles = _.filter(network.terminalRoles, function(r){return r.nodeId === nodeId;});

    // initialize a position if undefined.
    var margin = 100;
    var offset = 10;
    nodeobj.pos = nodeobj.pos || [margin + $("#sheets").offset().left, margin + $("#sheets").offset().top];

    // move the node if another node exists at the same place
    // limit to 1000 loop (I dont use infinite loop for safety)
    var rejectfunc = function(n){return n.id === nodeobj.id;};
    var findfunc  = function(n){return n.pos[0] === nodeobj.pos[0] && n.pos[1] === nodeobj.pos[1];};
    for (var i = 0; i < 1000; i++){
        if (_.chain(network.nodes).reject(rejectfunc).find(findfunc).value()){
            nodeobj.pos[0] += offset;
            nodeobj.pos[1] += offset;
        } else {
            break;
        }
    }

    // make sure position % 10 = 0
    nodeobj.pos[0] = truncate(nodeobj.pos[0]);
    nodeobj.pos[1] = truncate(nodeobj.pos[1]);

    //add to HTML structure
    $("#sheet_" + sheetname).append(
        _.template($("#nodeTemplate").html(), {node: nodeobj, delim: delim}));
    nodedom = $("#" + nodeId);

    nodedom.offset({left: nodeobj.pos[0], top: nodeobj.pos[1]})
           .css("display", "none");
    nodedom.children(".interminals").html(
        _.template($("#nodeIOTemplate").html(), {node: nodeobj, type: "in", term: existingTerminalRoles})
    );
    nodedom.children(".outterminals").html(
        _.template($("#nodeIOTemplate").html(), {node: nodeobj, type: "out", term: existingTerminalRoles})
    );

    var html = "";
    if (nodeobj.parameters.length === 0){
        html = "<tr><td>No parameters exist.</td></tr>";
    } else {
        _.each(nodeobj.parameters, function(parameter){
            html += _.template($("#nodeParameterTemplate").html(), {parameter: parameter});
        });
    }
    $("#" + nodeId + delim + "parametersForm table").html(html);

    if (noAnimation){
        nodedom.addClass("nodeSize" + userPreference.get("zoom"))
               .css("display", "block");
    } else {
        nodedom.addClass("nodeSize" + userPreference.get("zoom"))
               .fadeIn();
    }

    // register event callbacks
    nodedom.draggable({
        cancel: ".terminalcircle",
        containment: "#canvas_" + sheetname,
        scroll: true,
        grid: [10, 10],
        opacity: 0.8,
        drag: function(event, ui){
            var clickedNodeId = $(this).attr("id");
            var prevPos = _.findObject(network.nodes, "id", clickedNodeId).pos;
            var currPos = ui.offset;
            var diff = {top:  prevPos[1] - currPos.top,
                        left: prevPos[0] - currPos.left};
            var that = this;
            _.each(network.nodes, function(n){
                if (n.id === $(that).attr("id")){
                    n.pos = [truncate(currPos.left), truncate(currPos.top)];
                }
            });

            _.chain(network.nodes)
             .filter(function(n){return $("#" + n.id).hasClass("nodeSelected");})
             .reject(function(n){return n.id === clickedNodeId;})
             .each(function(n){
                n.pos[0] = truncate(n.pos[0] - diff.left);
                n.pos[1] = truncate(n.pos[1] - diff.top);
                console.log(n.pos[0], n.pos[1]);
                $("#" + n.id).offset({left: n.pos[0], top: n.pos[1]});
             });

            clearCanvas();
            drawLinks();
        },
        start: function(){
            $("#nodelist").unbind("keydown");
            if (!$(this).hasClass("nodeSelected")){
                $(".nodeSelected").removeClass("nodeSelected");
                $(this).addClass("nodeSelected");
            }
        }
    });

    nodeparamdom = $("#" + nodeId + delim + "parameters");
    nodeparamdom.dialog({
        autoOpen: false,
        show: "fade",
        hide: "fade",
        isExpanded: false,
        // height: 0.7 * ($("body").height() - $("#topmenu").height()),
        maxHeight: 0.9 * ($("body").height() - $("#topmenu").height()),
        buttons : {
            OK: function(){updateParams($(this)); $(this).dialog("close");},
            Cancel: function(){
                var tgtId = getBeforeDelim($(this).attr("id"));
                var prevParam = jQuery.data(this, "prevParam");
                _.each(network.nodes, function(n){
                    if (n.id === tgtId){
                        n.parameters = prevParam;
                    }
                });
                $(this).dialog("close");
            }
        },
        open: function(){
            var nodeId = getBeforeDelim($(this).attr("id"));
            var nodeobj = _.findObject(network.nodes, "id", nodeId);

            jQuery.data(this, "prevParam", JSON.parse(JSON.stringify(nodeobj.parameters)));

            var html = "";
            if (nodeobj.parameters.length === 0){
                html = "<tr><td>No parameters exist.</td></tr>";
            } else {
                _.each(nodeobj.parameters, function(parameter){
                    html += _.template($("#nodeParameterTemplate").html(), {parameter: parameter});
                });
            }
            $(this).find("table").html(html);
            if (!this.isExpanded){
                $(this).dialog("option", "width", nodeparamdom.find("table").width() + 100);
                this.isExpanded = true;
            }
            console.log(nodeparamdom.children(".type"));
            _.each($("#" + nodeId + delim + "parameters .type select"), function(type){
                if ($(type).val() !== "subnet_param"){$(type).trigger("change");}
            });
            _.each($("#" + nodeId + delim + "parameters .value select"), function(v){
                $(v).trigger("change");
            });
            _.each($("#" + nodeId + delim + "parameters value input"), function(v){
                $(v).trigger("change");
            });
        }
    });
    $("#" + nodeId + delim + "parametersForm").submit(function(event){
        updateParams($(this));
        nodeparamdom.dialog("close");
        event.preventDefault();
        event.stopPropagation();
    });

    // force value validation
    nodeparamdom.children(".value select").trigger("change");
    nodeparamdom.tooltip({position: {my: 'left center', at: 'right+5 center'}});

}

function setupParameterValidation(){
    $(document).on("change keyup", ".parameters input", function(event){
        var nodeId = getBeforeDelim($(this).parents("form").attr("id"));
        var paramname = $($(this).parent().siblings()[0]).children("label").text().trim();

        var nodeobj = _.findObject(network.nodes, "id", nodeId);
        var paramobj = JSON.parse(JSON.stringify(_.findObject(nodeobj.parameters, "name", paramname)));
        var oldparam = JSON.parse(JSON.stringify(_.findObject(nodeobj.parameters, "name", paramname)));
        paramobj.value = $(this).val();

        var result = testParameter(paramobj);
        if (result){
            $(this).addClass("error");
            raise(result, paramobj);
        } else {
            $(this).removeClass("error");
        }
    });

    // for list-type parameter values. Check the conditions and hide/show the other parameters.
    $(document).on("change", ".parameters .value select", function(event){
        // get useful info
        var nodeId = getBeforeDelim($(this).parents("form").attr("id"));
        var nodeobj = _.findObject(network.nodes, "id", nodeId);

        // update parameters
        var paramName = $($(this).parent().siblings()[0]).children("label").text().trim();
        _.findObject(nodeobj.parameters, "name", paramName).value = $(this).val();

        // control display = none/table-row
        var objdompair = _.map(nodeobj.parameters, function(p){
            return [p, _.find($("#" + nodeId + delim + "parametersForm tr"), function(tr){
                return $(tr).children(".name").text() === p.name;
            })];
        });

        _.chain(objdompair)
         .filter(function(pair){return pair[0].condition !== "";})
         .each(function(pair){
            var obj = pair[0];
            var dom = $(pair[1]);

            if (!obj){
                alert("ERROR: obj is undefined this should not happen");
                return;
            }
            if (!obj.condition){
                alert("ERROR: obj.condition is undefined this should not happen" + JSON.stringify(obj));
                return;
            }

            var condname = obj.condition.split("=")[0];
            var condvalue = obj.condition.split("=")[1].split(":");

            // hide parameter if condition is not satisfied
            var paramobj = _.findObject(nodeobj.parameters, "name", condname);
            if (paramobj && condvalue.indexOf(paramobj.value) === -1){
                dom.css("display", "none");
            } else {
                dom.css("display", "table-row");
            }
        });
    });
}

function updateParams(formobj){
    var nodeId = getBeforeDelim(formobj.attr("id"));
    var nodeobj = _.findObject(network.nodes, "id", nodeId);
    _.each(nodeobj.parameters, function(p){
        p.value = $("#" + nodeId + delim + "parametersForm [name='" + p.name + "']").val();
        if (p.type === "file"){
            p.type = "string";
        }
    });
    subnetParamMaintain(nodeId);
    updateDynamicNodeDOM(network, true);
}
function removeNode(nodeId, DOMonly, noAnimation, calledFromUpdateDynamicNodeDom){
    if (!DOMonly){
        _.chain(network.terminalRoles)
         .filter(function(r){return r.nodeId === nodeId;})
         .each(function(r){
            updateTerminalRoles(r, "del");
        });

        network.nodes = _.reject(network.nodes, function(node){
            return nodeId === node.id;
        });

        network.links = _.reject(network.links, function(link){
            return nodeId === link.fromNodeId |
                   nodeId === link.toNodeId;
        });
    }

    function _destroy(nodeId){

        //remove node DOM
        $("#" + nodeId).draggable("destroy");
        _.chain($(".node")).filter(function(n){return $(n).attr("id") === nodeId;})
         .each(function(n){$(n).unbind().remove();});

        //remove parameter dialog
        $("#" + nodeId + delim + "parameters").dialog("destroy");
        $("#" + nodeId + delim + "parameters").unbind().remove();

        removeNonExistingSubnetParam();

        if (!DOMonly){
            clearCanvas();
            drawLinks();
        }
    }
    if (!noAnimation && $("#" + nodeId).css("display") === "block"){
        $("#" + nodeId).fadeOut(function(){_destroy(nodeId);});
        adjustScroll();
    } else {
        _destroy(nodeId);
    }
}

function setupNodeEvents(){
    // parameter change
    $(document).on("change", ".parameters td.type select", function(event){
        var nodeId = getBeforeDelim($($(this).parents("div")[0]).attr("id"));
        var paramName = $(this).attr("name").replace("_type", "");

        var nodeobj = _.findObject(network.nodes, "id", nodeId);
        var paramobj = _.findObject(nodeobj.parameters, "name", paramName);
        var newtype = $(this).val();
        var oldtype = paramobj.type;
        var oldvalue = paramobj.value;

        // if no types changed, just quit.
        paramobj.type = newtype;
        if (newtype === "subnet_param"){
            paramobj.value = paramName;
        }

        // update parameter
        var nodeIndex = _.indexOf(network.nodes, nodeobj);
        var paramIndex = _.indexOf(nodeobj.parameters, paramobj);
        network.nodes[nodeIndex].parameters[paramIndex] = JSON.parse(JSON.stringify(paramobj));

        // // rendering: keep old values
        // var renderParams = JSON.parse(JSON.stringify(nodeobj.parameters));
        // var funcRenderParams = function(formValue){
        //     _.find(renderParams, function(p){
        //         if (p.name === $(formValue).attr("name"))
        //             p.value = $(formValue).val();
        //     });
        // };
        // _.each($(".paramDialog .value input"), funcRenderParams);
        // _.each($(".paramDialog .value select"), funcRenderParams);
        // renderParams[paramIndex] = JSON.parse(JSON.stringify(paramobj));
        $(this).parentsUntil("tr").parent().html(
            _.template($("#nodeParameterTemplate").html(), {parameter: paramobj})
             .replace("<tr>", "").replace("</tr>", "")
        );

        // if type === select, update options
        if (newtype === "file"){
            var node = $("#" + nodeId + delim + "parameters" + " .value select[name=" + paramName + "]");
            node.html("");
            _.chain(fileCollection.toJSON())
             .groupBy("type").map(function(gr){return _.sortBy(gr, "name");}).flatten()
             .each(function(f){
                var name = f.name;
                if (oldvalue === name){
                    node.append("<option value='" + name + "' selected>" + name + "</option>");
                } else {
                    node.append("<option value='" + name + "'>" + name + "</option>");
                }
            });
            if (fileCollection.toJSON().length === 0){
                node.append("<option value='no files found' selected>no files found</option>");
            }
         }

        $(".parameters .value select").trigger("change");
    });

    // add node from menu
    $(document).on("click", ".nodeElement", function(event){
        addNode($(this).text(), currentTab);
    });

    $(document).on("mousedown", ".terminal", function(event){
        if (currentLinkFrom === false){
            currentLinkFrom = $(this);
        }
        isDraggingStarted = false;
        $("#sheet_" + currentTab).selectable("destroy");
        event.stopPropagation();
    });
    $(document).on("mouseup", "body", function(event){
        currentLinkFrom = false;
        setupSelectable(currentTab);
        clearCanvas();
        drawLinks();
    });
    $(document).on("mouseup", ".terminal", function(event){
        if (currentLinkFrom === false){
            return;
        }
        if (isDraggingStarted === false){
            return;
        }

        // connect a link
        var fromTerminal = (currentLinkFrom.parent().hasClass("interminals")) ? "interminals" : "outterminals";
        if (fromTerminal === "outterminals"){
            newLink({
                fromNodeId: currentLinkFrom.parents(".node").attr("id"),
                fromTermName: currentLinkFrom.attr("name"),
                toNodeId: $(this).parents(".node").attr("id"),
                toTermName: $(this).attr("name"),
                sheet: currentTab
            });
        } else {
            newLink({
                toNodeId: currentLinkFrom.parents(".node").attr("id"),
                toTermName: currentLinkFrom.attr("name"),
                fromNodeId: $(this).parents(".node").attr("id"),
                fromTermName: $(this).attr("name"),
                sheet: currentTab
            });
        }
        currentLinkFrom = false;
        isDraggingStarted = false;
        clearCanvas();
        drawLinks();
    });

    $(document).on("mousemove", "canvas", function(event){
        if (currentLinkFrom !== false){
            clearCanvas();
            drawLinks();
            drawCandidateLinks();
            drawCurrentLine(null, event);
            isDraggingStarted = true;
        }
    });

    $(document).on("dblclick", ".node", function(event){
        var clickedNodeId = $(this).attr("id");
        $("#" + clickedNodeId + delim + "parameters").dialog("open");
        event.preventDefault();
        event.stopPropagation();
    });

    $(document).on("click", ".node", function(event){
        if (selectedNodeObj){
            $("#" + selectedNodeObj.id).removeClass("nodeSelected");
        }
        selectedNodeObj = _.findObject(network.nodes, "id", $(this).attr("id"));
        $(this).addClass("nodeSelected");
        event.preventDefault();
        event.stopPropagation();
    });

    // context menu for node
    var selected = "";
    $("#sheets").contextMenu({
        selector: ".sheetTab",
        build: function($trigger, event){
            var sheetname_ = $trigger.find("a").attr("href").replace("#sheet_", "");
            var types = ["Change the name", "Change to subnet", "Change to iterator", "Delete the sheet"];
            var items = {};
            _.each(types, function(type){
                items[type] = {name: type, icon: "edit"};
            });
            items["Delete the sheet"] = {name: "Delete the sheet", icon: "delete"};

            _.each(_.keys(items), function(key){
                if (key.indexOf(network.sheets[sheetname_].type) !== -1){
                    delete items[key];
                }
            });

            if ($($trigger).attr("aria-controls") === "sheet_MAIN"){
                return {};
            } else {
                return {
                    callback: function(key, options){
                        var sheetname = $trigger.find("a").attr("href");
                        sheetname = sheetname.substr(sheetname.indexOf("#sheet_") + "#sheet_".length);
                        if (key === "Change the name"){
                            selected = sheetname;
                            $("#sheetTabDialog").dialog("open");
                        } else if (key === "Delete the sheet"){
                           if (confirm("Are you sure to delete " + sheetname + "?")){
                               removeSheet(sheetname);
                           }
                        } else {
                            network.sheets[sheetname].type = key.replace("Change to ", "");
                            $(".sheetTab[aria-controls='sheet_" + sheetname + "']").attr(
                                "title", 'Sheet type: ' + network.sheets[sheetname].type);
                            $(".sheetTab[aria-controls='sheet_" + sheetname + "'] a")
                                .removeClass("sheetType_iterator")
                                .removeClass("sheetType_subnet")
                                .addClass("sheetType_" + network.sheets[sheetname].type);
                            }
                    },
                    items: items
                };
            }
        }
    });

    $("#sheetTabDialogForm").submit(function(event){
        $($("#sheetTabDialog").siblings(".ui-dialog-buttonpane").find(".ui-button")[0]).trigger("click");
        event.preventDefault();
        event.stopPropagation();
    });
    $("#sheetTabDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons :  {
            "OK": function(){
                newname = $(this).find("input").val();
                if (newname === ""){
                    app.set("execstat", "Error sheet name cannot be empty");
                    return;
                }

                if (_.keys(network.sheets).indexOf(newname) !== -1){
                    app.set("execstat", "Error: Sheet name is duplicated");
                    return;
                }

                // rename node.sheet and dynamic nodes
                _.each(network.nodes, function(n){if (n.sheet === selected){n.sheet = newname;}});
                _.each(network.nodes, function(n){
                    if (n.name === selected){
                        var oldid = n.id;
                        n.id = "node_" + newname + "_" + n.index;
                        n.name = newname;

                        $("#" + oldid + " .nodename").text(newname);
                        $("#" + oldid + " .nodeid").text(n.id);
                        _.each($("#" + oldid + " .terminal"), function(t){
                            $(t).attr("id", n.id + delim + getAfterDelim($(t).attr("id")));
                        });
                        _.each($("[aria-describedby=" + oldid + delim + "parameters]"), function(d){
                            $(d).attr("aria-describedby", n.id + delim + "parameters");
                            $(d).find(".ui-dialog-title").text("Parameters of " + n.id);
                            $(d).find("#" + oldid + delim + "parameters").attr("id", n.id + delim + "parameters");
                            $(d).find("#" + oldid + delim + "parametersForm").attr("id", n.id + delim + "parametersForm");
                        });
                        $("[aria-controls=sheet_" + selected + "]").attr("aria-controls", "sheet_" + newname);
                        $("#" + oldid).attr("id", n.id);
                        $("#" + oldid).draggable("refresh");
                    }
                });

                //rename links
                _.each(network.links, function(l){if (l.sheet === selected){l.sheet = newname;}});
                _.each(network.links, function(l){
                    var regexp = "node_" + selected;
                    if (l.fromNodeId.match(regexp)){
                        l.fromNodeId = "node_" + newname + "_" + l.fromNodeId.split("_").slice( -1 );
                    }
                    if (l.toNodeId.match(regexp)){
                        l.toNodeId = "node_" + newname + "_" + l.toNodeId.split("_").slice( -1 );
                    }
                });

                //rename terminalRoles
                _.each(network.terminalRoles, function(l){if (l.sheet === selected){l.sheet = newname;}});
                _.each(network.terminalRoles, function(l){
                    var regexp = "node_" + selected;
                    if (l.nodeId.match(regexp)){
                        l.nodeId = "node_" + newname + "_" + l.nodeId.split("_").slice( -1 );
                    }
                });

                //rename Sheets
                network.sheets[newname] = {
                    name: newname,
                    type: network.sheets[selected].type
                };
                delete network.sheets[selected];

                //rename nodedefs
                _.find(nodedefs,function(d){return d.name === selected;}).name = newname;

                $("#sheet_" + selected).attr("id", "sheet_" + newname);
                $("#canvas_" + selected).attr("id", "canvas_" + newname);
                $("a[href=#sheet_" + selected + "]").text(newname);
                $("a[href=#sheet_" + selected + "]").attr("href", "#sheet_" + newname);
                if (currentTab === selected) {currentTab = newname;}
                $("#sheets").tabs("refresh");

                updateNodeList();
                sheetsView.toggleMinimize();
                $(this).dialog("close");

            },
            "Cancel": function(){
                $(this).dialog("close");
            }
        }
    });

    var themeSelectorItems = {
        "theme" : {name: "Set theme",
                 items: {
                    "theme-Fire": {"name": "Fire"},
                    "theme-Sea" : {"name": "Sea" },
                    "theme-Dark": {"name": "Dark"},
                    "theme-Classic": {"name": "Classic"}
                 }
             }
        };
    var themeSelectorCallback = function(key, options, clickedNodeId){
        switch (key){
            case "theme-Fire":
            case "theme-Sea":
            case "theme-Dark":
            case "theme-Classic":
                var theme = key.slice(6);
                var nodes = (clickedNodeId) ? $("#" + clickedNodeId) : $(".nodeSelected");
                _.each(nodes, function(n){
                    $(n).removeClass("nodeThemeFire");
                    $(n).removeClass("nodeThemeSea");
                    $(n).removeClass("nodeThemeDark");
                    $(n).removeClass("nodeThemeClassic");
                    $(n).addClass("nodeTheme" + theme);
                    _.each(network.nodes, function(n2){
                        if (n2.id === $(n).attr("id")){
                            n2.theme = theme;
                        }
                    });
                });
            break;
        }
    };
    // context menu for canvas
    $("#sheets").contextMenu({
        selector: ".sheetContainer canvas",
        build: function($trigger, event){
            if ($(".nodeSelected").length > 0){
                return {
                    callback: themeSelectorCallback,
                    items: themeSelectorItems
                };
            }
        }
    });
    // context menu for nodes
    $("#sheets").contextMenu({
        selector: ".node",
        callback: function(key, options){
            var clickedNodeId = $(this).attr("id");
            var clickedNodeObj = _.findObject(network.nodes, "id", clickedNodeId);
            var clickedNodeName = nodeId2nodename(clickedNodeId);
            var clickedNodeCat  = nodeId2category(clickedNodeId);

            switch (key){
                case "delete":
                removeNode(clickedNodeId);
                break;

                case "property":
                $("#" + clickedNodeId + delim + "parameters").dialog("open");
                break;

                case "showHelp":
                if (clickedNodeCat.indexOf("HARK") !== -1){
                    var urltmpl = "http://winnie.kuis.kyoto-u.ac.jp/HARK/document/1.2.0/hark-document-LANG/subsec-REPLACE.html";
                    window.open(urltmpl.replace("REPLACE", clickedNodeName)
                                       .replace("LANG", userPreference.get("lang")),
                                "Help: " + clickedNodeName);
                } else {
                    alert("No help for this node.");
                }
                break;

                case "addInput":
                $("#AddIODialog").attr("clickedNodeId", clickedNodeId)
                                 .attr("title", "Add Input")
                                 .dialog("option", "title", "Add Input")
                                 .dialog("open");
                break;

                case "addOutput":
                $("#AddIODialog").attr("clickedNodeId", clickedNodeId)
                                 .attr("title", "Add Output")
                                 .dialog("option", "title", "Add Output")
                                 .dialog("open");
                break;
            }
            themeSelectorCallback(key, options, clickedNodeId);
        },
        items: _.extend({
            "property" : {name: "Property", icon: "edit"},
            "delete":  {name: "Delete", icon: "delete"},
            "addInput":  {name: "Add Input", icon: "add"},
            "addOutput":  {name: "Add Output", icon: "add"},
            "showHelp" : {name: "Help", icon: "help"}
        }, themeSelectorItems)
    });
    console.log(_.extend({
            "property" : {name: "Property", icon: "edit"},
            "delete":  {name: "Delete", icon: "delete"},
            "addInput":  {name: "Add Input", icon: "add"},
            "addOutput":  {name: "Add Output", icon: "add"},
            "showHelp" : {name: "Help", icon: "help"}
        }, themeSelectorItems), themeSelectorItems);

    //context menu for input terminals
    $("#sheets").contextMenu({
        selector: ".terminal",
        trigger: isSelenium,
        build: function($trigger, event){
            var items = {};
            var termid = $trigger.attr("id");
            var terminalSide = ($trigger.parent().hasClass("interminals")) ? "Input" : "Output";
            var termName = getAfterDelim(termid);
            var nodeobj = _.findObject(network.nodes, "id", getBeforeDelim(termid));

            // if the terminal is already input, set "unset" menu
            // otherwise, set "set as input" menu
            var isIOAlreadySet = _.find(network.terminalRoles, function(role){
                return role.role !== "condition" && role.nodeId + delim + role.termName === termid;});
            if (isIOAlreadySet){
                items.unset = {name: "Delete " + terminalSide, icon: "delete"};
            } else {
                items.set = {name: "Set as " + terminalSide, icon: "add"};
            }

            var isLinkExists = _.find(network.links, function(link){
                return termid === link.fromNodeId + delim + link.fromTermName ||
                       termid === link.toNodeId   + delim + link.toTermName;
                   });
            if (isLinkExists){items.unlink = {name: "Delete the link", icon: "delete"};}

            var dsc = _.findObject(nodeobj[terminalSide.toLowerCase() + "s"],
                                     "name", termName).description;
            if (dsc && dsc.indexOf("added by the user") !== -1 || dsc.indexOf("Dynamically added") !== -1){
                items.deleteIO = {name: "Delete this " + terminalSide, icon: "delete"};
            }

            if (terminalSide == "Output"){
                var isConditionAlreadySet = _.find(
                    network.terminalRoles,
                    function(role){
                        return role.role === "condition" && role.nodeId + delim + role.termName == termid;
                    });
                if (isConditionAlreadySet){
                    items.deleteCondition = {name: "Delete Condition", icond: "delete"};
                } else {
                    items.setCondition    = {name: "Set as Condition", icon: "add"};
                }
            }

            return {
                callback:
                function (key, options){
                    var clickedTermId = $(this).attr("id");
                    var clickedNodeId = getBeforeDelim(clickedTermId);
                    var clickedNodeObj = _.findObject(network.nodes, "id", clickedNodeId);
                    var clickedNodeName = nodeId2nodename(clickedNodeId);
                    var clickedTermName = getAfterDelim(clickedTermId);
                    var delTarget;
                    switch (key){
                        case "unset":
                            delTarget = _.find(network.terminalRoles,
                                function(role){return role.nodeId + delim + role.termName === termid && role.role !== "condition";});
                            updateTerminalRoles(delTarget, "del");

                            if (currentTab !== "MAIN"){
                                var dynamic = _.findObject(network.nodes, "name", currentTab);
                                network.links = _.reject(network.links,
                                    function(link){return link.fromNodeId === dynamic.id && link.fromTermName === clickedTermName;});
                                network.links = _.reject(network.links,
                                    function(link){return link.toNodeId === dynamic.id && link.toTermName === clickedTermName;});
                            }
                            break;
                        case "set":
                            var ioobj = (terminalSide === "Input") ? clickedNodeObj.inputs : clickedNodeObj.outputs;
                            ioobj = _.findObject(ioobj, "name", clickedTermName);
                            $("#terminalRoleNameDialog")
                                .attr("clickedNodeId", clickedNodeId)
                                .attr("clickedTermName", clickedTermName)
                                .attr("role", terminalSide.toLowerCase())
                                .attr("ioobj_type", ioobj.type)
                                .attr("ioobj_description", ioobj.description);

                            var numNames = _.chain(network.terminalRoles)
                                .filter(function(r){return r.sheet === currentTab;})
                                .filter(function(r){return r.name.replace(/_[0-9]$/, "") === clickedTermName;})
                                .value();
                            if (numNames.length === 0){
                                $("#terminalRoleName").val(clickedTermName);
                            } else {
                                var filterfunc = function(r){return r.sheet === currentTab;};
                                var findfunc   = function(r){return r.name === candidate;};
                                for (var i = 0; i < numNames.length + 1; i++){
                                    var candidate = (i === 0) ? clickedTermName : (clickedTermName + "_" + i);
                                    if (_.chain(network.terminalRoles)
                                        .filter(filterfunc)
                                        .find(findfunc).value() === undefined){
                                        $("#terminalRoleName").val(candidate);
                                        break;
                                    }
                                }
                            }
                            $("#terminalRoleNameDialog").dialog("open");

                            break;
                        case "unlink":
                            network.links = _.reject(network.links, function(link){
                                return clickedTermId === link.fromNodeId + delim + link.fromTermName ||
                                       clickedTermId === link.toNodeId   + delim + link.toTermName;
                            });
                            clearCanvas();
                            drawLinks();
                            break;
                        case "deleteCondition":
                            delTarget = _.find(network.terminalRoles,
                                function(role){return role.nodeId + delim + role.termName === termid;});
                            updateTerminalRoles(delTarget, "del");
                            break;
                        case "setCondition":
                            newTerminalRole({
                                nodeId: clickedNodeId,
                                termName: clickedTermName,
                                name: "CONDITION",
                                role: "condition",
                                sheet: currentTab
                            });
                            break;
                        case "deleteIO":
                            clickedNodeObj[terminalSide.toLowerCase() + "s"] = _.reject(
                                clickedNodeObj[terminalSide.toLowerCase() + "s"], function(t){
                                return t.name === clickedTermName;
                            });
                            network.terminalRoles = _.reject(network.terminalRoles, function(r){
                                return r.nodeId === clickedNodeId &&
                                       r.termName === clickedTermName;
                            });

                            $("#" + clickedTermId).remove();
                        break;

                    }
                    updateDynamicNodeDOM(network);
                    clearCanvas();
                    drawLinks();
                },
                items: items};
        }
    });
}

function nodeId2category(nodeId){
    var nodename = nodeId.substring(nodeId.indexOf("_") + 1,
                                    nodeId.lastIndexOf("_"));
    var def = _.findObject(nodedefs, "name", nodename);
    if (def){
        return def.category;
    } else {
        return "Undefined";
    }
}

function updateTerminalRoles(role, command, stopUpdateDOM){
    // command in ["add", "del"]

    var obj;
    if (command === "add"){
        if (role.sheet !== "MAIN"){
            var newio = {
                name: role.name,
                type: "any",
                description: "Dynamic"
            };
            updateDynamicNode(role.sheet, role.role, newio);

            if (role.role !== "condition"){
                // add to defs
                _.chain(nodedefs)
                 .filter(function(def){return def.name === role.sheet;})
                 .each(function(def){
                    if (_.find(def[role.role + "s"], function(io){return io.name === role.name;}) === undefined){
                        def[role.role + "s"].push(newio);
                    }
                    });

                 // add to nodes
                 _.chain(network.nodes)
                  .filter(function(n){return n.name === role.sheet;})
                  .each(function(n){
                    if (_.find(n[role.role + "s"], function(io){return io.name === role.name;}) === undefined){
                        n[role.role + "s"].push(newio);
                    }
                  });
            }
        }
        obj = $("#" + role.nodeId + delim + role.termName + " .terminalRoleName");
        if (obj.text().trim().replace(/[\W]+/g, " ").split(" ").indexOf(role.name) === -1){
            obj.text(obj.text() + " " + role.name);
        }
    } else if (command === "del"){
        if (role.role !== "condition"){
            // remove from defs
            _.chain(nodedefs)
             .filter(function(def){return def.name === role.sheet;})
             .each(function(def){
                def.outputs = _.reject(def.outputs, function(o){return o.name === role.termName;});
                def.inputs  = _.reject(def.inputs,  function(i){return i.name === role.termName;});
             });

            // remove from nodes
            _.chain(network.nodes)
             .filter(function(n){return n.name === role.sheet;})
             .each(function(n){
                n.outputs = _.reject(n.outputs, function(o){return o.name === role.name;});
                n.inputs  = _.reject(n.inputs,  function(i){return i.name === role.name;});
            });

            // remove terminalRoles
            tmp = _.chain(network.terminalRoles)
                   .reject(function(r){
                    if (nodeId2nodename(r.nodeId) === role.sheet && r.termName === role.name){
                        updateTerminalRoles(r, "del");
                        return true;
                    } else {
                        return false;
                    }
                }).value();
            tmp = network.terminalRoles;

            network.links = _.reject(network.links, function(l){
                return (nodeId2nodename(l.fromNodeId) === role.sheet && l.fromTermName === role.name) ||
                       (nodeId2nodename(l.toNodeId)   === role.sheet && l.toTermName   === role.name);
            });
            // _.each(network.terminalRoles, function(r){
                // if (nodeId2nodename(r.nodeId) === role.sheet && r.termName === role.name)
                    // updateTerminalRoles(r, "del");
              // });
        }

        //remove from terminalRoles
        network.terminalRoles = _.reject(network.terminalRoles, function(r){
            return r === role;
        });
        obj = $("#" + role.nodeId + delim + role.termName + " .terminalRoleName");
        obj.text(obj.text().replace(role.name, ""));
    } else {
        console.log("updateTerminalRoles: unknown command: " + command + ". Command must be add or del.");
    }

    if (stopUpdateDOM !== true || role.sheet === "MAIN"){
        updateDynamicNodeDOM(network);
    }
}

function updateDynamicNodeDOM(networkobj, forceClose){
    // if (role && command){
    //     console.log(role, command);
    //     if (command === "add" && role.role !== "condition"){
    //         var selector = "#" + role.nodeId + " " + role.role.replace("put", "") + "terminals";
    //         var tmpl = '<die class="terminal" description="<%- description %>" name="<%= name %>" type="<%- object_type %>" id="<%= nodeId %>' + delim + '<%= name %>" ><div class="terminalcircle"><span class="terminalRoleName"></span></div><%= name %></div>';
    //         console.log(selector);
    //         console.log(_.template(tmpl, role));
    //         $(selector).append(_.template(tmpl, role));
    //     }
    //     else if (command === "del" && role.role !== "condition"){

    //     }
    //     return;
    // }

    // remove and add
    var updatedNodeNames = _.chain(networkobj.nodes)
                            .groupBy("sheet")
                            .keys()
                            .reject(function(k){return k === "MAIN";})
                            .value();
    _.chain(networkobj.nodes)
     .filter(function(n){return updatedNodeNames.indexOf(n.name) !== -1;})
     .each(function(n){
        var sel = "#" + n.id + delim + "parameters";
        var isOpen = $(sel).dialog("isOpen");
        removeNode(n.id, true, true, true, true);
        addNodeDOM(n, true);
        if (isOpen && !forceClose){
            $(sel).dialog("open");
        }

        _.chain(networkobj.terminalRoles)
         .filter(function(r){return r.nodeId === n.id;})
         .each(function(r){
            var obj = $("#" + r.nodeId + delim + r.termName + " .terminalRoleName");
            if (obj.text().trim().replace(/[\W]+/g, " ").split(" ").indexOf(r.name) === -1){
                obj.text(obj.text() + " " + r.name);
            }
         });
     });
}

/* drawing functions.*/
function clearCanvas(canvasid){
    canvasid = canvasid || "canvas_" + currentTab;
    var curcanvas = document.getElementById(canvasid);
    var ctx = curcanvas.getContext("2d");
    ctx.fillStyle = "#f8f8f8";
    ctx.fillRect(0, 0, curcanvas.width, curcanvas.height);
}

function truncate(pos){
    return parseInt(pos / 10, 10) * 10.0;
}

function Bezier(poses, context){
/*http://ruiueyama.tumblr.com/post/11197882224*/
    function interp(a, b, f){return a + f * (b - a);}
    var split = 15;

    context.moveTo(poses[0].left, poses[0].top);
    for (var i = 0; i < split; i++){
        var p1 = {left: interp(poses[0].left, poses[1].left, 1.0 * i / split),
                  top : interp(poses[0].top,  poses[1].top,  1.0 * i / split)};
        var p2 = {left: interp(poses[1].left, poses[2].left, 1.0 * i / split),
                  top : interp(poses[1].top,  poses[2].top,  1.0 * i / split)};

        context.lineTo(interp(p1.left, p2.left, 1.0 * i / split),
                       interp(p1.top,  p2.top,  1.0 * i / split));
    }
    context.lineTo(poses[2].left, poses[2].top);
}

function intelligentLink(frompos, topos, context, mtop, mleft){
    mode = userPreference.get("link");
    mleft = mleft || (frompos.left + topos.left) / 2.0;
    mtop  = mtop || (frompos.top + topos.top) / 2.0;
    var margin = Math.min(Math.abs(frompos.left - topos.left), 20);
    var poses = [
        frompos,
        {left: frompos.left + margin, top: frompos.top},
        {left: frompos.left + margin, top: (frompos.top + mtop) / 2.0},
        {left: frompos.left + margin, top: mtop},
        {left: mleft,                 top: mtop},
        {left: topos.left - margin,   top: mtop},
        {left: topos.left - margin,   top: (topos.top + mtop) / 2.0},
        {left: topos.left - margin,   top: topos.top},
        topos
    ];

    if (mode === "Bezier"){
        if (frompos.left < topos.left){
            var m1 = {left: mleft, top: frompos.top};
            var m2 = {left: mleft, top: mtop};
            var m3 = {left: mleft, top: topos.top};

            Bezier([frompos, m1, m2], context);
            Bezier([m2, m3, topos], context);
        } else {
            Bezier([poses[0], poses[1], poses[2]], context);
            Bezier([poses[2], poses[3], poses[4]], context);
            Bezier([poses[4], poses[5], poses[6]], context);
            Bezier([poses[6], poses[7], poses[8]], context);
        }
    } else if (mode === "Lines"){
        if (frompos.left < topos.left){
            context.moveTo(frompos.left, frompos.top);
            context.lineTo(mleft, frompos.top);
            context.lineTo(mleft, topos.top);
            context.lineTo(topos.left, topos.top);
        } else {
            context.moveTo(poses[0].left, poses[0].top);
            for (var i = 1; i < poses.length; i++){
                context.lineTo(poses[i].left, poses[i].top);
            }
        }
    }
}

function drawCandidateLinks(){
    var targetSide = (currentLinkFrom.parent().hasClass("interminals")) ? "outputs" : "inputs";
    var mySide = (currentLinkFrom.parent().hasClass("interminals")) ? "inputs" : "outputs";
    var curFromId = currentLinkFrom.attr("id");

    var currentLinkFromType = _.findObject(network.nodes, "id", getBeforeDelim(curFromId))[mySide];
    currentLinkFromType = _.findObject(currentLinkFromType,
                                       "name", getAfterDelim(curFromId)).type;
    var candidates = _.chain(network.nodes)
                        .filter(function(n){return n.sheet === currentTab;})
                        .map(function(n){
                            return _.chain(n[targetSide])
                                    .filter(function(term){return isTypesMatch(term.type, currentLinkFromType);})
                                    .map(function(term){
                                        if (mySide === "outputs"){
                                            return {
                                                 fromNodeId: getBeforeDelim(curFromId),
                                                 fromTermName: getAfterDelim(curFromId),
                                                 toNodeId  : n.id,
                                                 toTermName: term.name,
                                                 sheet: currentTab
                                            };
                                        } else {
                                            return {
                                                 fromNodeId  : n.id,
                                                 fromTermName: term.name,
                                                 toNodeId: getBeforeDelim(curFromId),
                                                 toTermName: getAfterDelim(curFromId),
                                                 sheet: currentTab
                                            };
                                        }
                            }).value();
                        }).flatten().value();
    candidates = _.filter(candidates, function(c){return testLink(c) === "";});
    drawLinks(undefined, candidates, "#888", 1, true);
}

function drawLinks(canvasid, links, style, width, isDotted){
    links = links || network.links;
    style = style || "black";
    width = width || 2;
    canvasid = canvasid || "canvas_" + currentTab;

    var curcanvas = document.getElementById(canvasid);
    var context = curcanvas.getContext("2d");
    context.strokeStyle = style;
    context.lineWidth = width;
    if (isDotted){
        if (context.setLineDash) {context.setLineDash([3, 3]);}
        context.mozDash = [3, 3];
    } else {
        if (context.setLineDash) {context.setLineDash([1, 0]);}
        context.mozDash = [1, 0];
    }

    context.beginPath();
    _.each(links, function(link){
        tmpl = "#<%= id %><%= delim %><%= link %> .terminalcircle";
        var fromNodeId  = _.template(tmpl, {id: link.fromNodeId, link: link.fromTermName, delim: delim});
        var toNodeId    = _.template(tmpl, {id: link.toNodeId, link: link.toTermName, delim: delim});
        if ($(fromNodeId).length === 0){return;}
        if ($(toNodeId).length === 0){return;}
        if ($("#" + link.fromNodeId).length === 0){return;}
        if ($("#" + link.toNodeId).length === 0){return;}

        var frompos = {
            left: $(fromNodeId).offset().left -
                  $(fromNodeId).parentsUntil(".sheetContainer").parent().offset().left +
                  $(fromNodeId).parentsUntil(".sheetContainer").parent().scrollLeft() + 5,
            top:  $(fromNodeId).offset().top -
                  $(fromNodeId).parentsUntil(".sheetContainer").parent().offset().top +
                  $(fromNodeId).parentsUntil(".sheetContainer").parent().scrollTop() + 5
        };
        var topos   = {
            left: $(toNodeId).offset().left -
                  $(toNodeId).parentsUntil(".sheetContainer").parent().offset().left +
                  $(toNodeId).parentsUntil(".sheetContainer").parent().scrollLeft() + 5,
            top:  $(toNodeId).offset().top -
                  $(toNodeId).parentsUntil(".sheetContainer").parent().offset().top +
                  $(toNodeId).parentsUntil(".sheetContainer").parent().scrollTop() + 5
        };
        var fromnodepos = {
            left: $("#" + link.fromNodeId).offset().left -
                  $("#" + link.fromNodeId).parent().offset().left +
                  $("#" + link.fromNodeId).parent().scrollLeft() + 5,
            top:  $("#" + link.fromNodeId).offset().top -
                  $("#" + link.fromNodeId).parent().offset().top +
                  $("#" + link.fromNodeId).parent().scrollTop() + 5
        };
        fromnodepos.bottom = fromnodepos.top + $("#" + link.fromNodeId).height();
        var tonodepos   = {
            left: $("#" + link.toNodeId).offset().left -
                  $("#" + link.toNodeId).parent().offset().left +
                  $("#" + link.toNodeId).parent().scrollLeft() + 5,
            top:  $("#" + link.toNodeId).offset().top -
                  $("#" + link.toNodeId).parent().offset().top +
                  $("#" + link.toNodeId).parent().scrollTop() + 5
        };
        tonodepos.bottom = tonodepos.top + $("#" + link.toNodeId).height();

        var offset = -6; // magic number. probably related to padding or margin?
        if (fromnodepos.bottom < tonodepos.top){
            intelligentLink(frompos, topos, context,
                (fromnodepos.bottom + tonodepos.top) / 2.0 + offset);
        } else if (tonodepos.bottom < fromnodepos.top){
            intelligentLink(frompos, topos, context,
                (tonodepos.bottom + fromnodepos.top) / 2.0 + offset);
        } else {
            intelligentLink(frompos, topos, context);
        }
    });
    context.stroke();
    context.closePath();
}

function crossBrowserMousePos(event){
    /*http://www.jacklmoore.com/notes/mouse-position/*/
    var target = event.target || event.srcElement,
        rect = target.getBoundingClientRect(),
        offsetX = event.clientX - rect.left,
        offsetY = event.clientY - rect.top;
    return {left: offsetX, top: offsetY};
}
var debug;
function drawCurrentLine(canvasid, event){
    canvasid = canvasid || "canvas_" + currentTab;
    var curcanvas = document.getElementById(canvasid);
    var context = curcanvas.getContext("2d");

    var circle = currentLinkFrom.find(".terminalcircle");
    var parents = circle.parentsUntil(".sheetContainer").parent();

    var termpos = {
        left: circle.offset().left - parents.offset().left + parents.scrollLeft(),
        top : circle.offset().top  - parents.offset().top  + parents.scrollTop()
    };
    var mousepos = crossBrowserMousePos(event);

    context.strokeStyle = "black";
    context.beginPath();
    var mySide = (currentLinkFrom.parent().hasClass("interminals")) ? "inputs" : "outputs";
    if (mySide === "inputs"){
        intelligentLink(mousepos, termpos, context);
    } else {
        intelligentLink(termpos, mousepos, context);
    }
    context.stroke();
    context.closePath();
}

function relativePos(of, from){
    if (of === undefined || from === undefined){
        return undefined;
    }
    // of and from must be a jQuery object $(...)
    var pad = {};
    pad.top = from.css("padding-top").slice(0, -2);
    pad.left = from.css("padding-left").slice(0, -2);

    var mid = {};
    mid.top = of.css("height").substr(0, 2) / 2.0;
    mid.left = of.css("width").substr(0, 2) / 2.0;

    var pos = {};
    pos.top  = - from.position().top  + of.position().top - pad.top + mid.top;
    pos.left = - from.position().left + of.position().left - pad.left + mid.left;

    return pos;
}

/* setup #sheets */
function addDynamicNode(sheetname){
    nodedefs.push({
        name:        sheetname,
        category:    "Dynamic",
        description: "Dynamic node",
        inputs:      [],
        outputs:     [],
        parameters:  [],
        def:         "Dynamic",
        theme:       "Classic"
    });
    updateNodeList();
}

function updateDynamicNode(sheetname, type, option){
    // update inputs/outputs/parameters of dynamic node
    if (sheetname === "MAIN"){return;}

    var def = _.findObject(nodedefs, "name", sheetname);

    if (type === "input" || type === "output"){
        if (def[type + "s"] &&
           _.find(def[type + "s"], function(i){return i.name === option.name;}) === undefined){
            def[type + "s"].push(option);
        }
        _.each(network.nodes, function(n){
            if (n.name === sheetname &&
               _.find(n[type + "s"], function(i){return i.name === option.name;}) === undefined){
                n[type + "s"] = n[type + "s"] || [];
                n[type + "s"].push(option);
            }
        });
    }
}

function addSheet(sheetname, sheettype){
    if (newSheet({name: sheetname, type: sheettype})){
        addSheetDOM(network.sheets[sheetname]);
        addDynamicNode(sheetname);
    }
}

function getSheetId(sheetname){
    return _.indexOf($("#sheets .ui-tabs-nav li[role=tab]"), function(tab){
        return $(tab).attr("aria-controls") === "sheet_" + sheetname;
    });
}

function addSheetDOM(sheet){
    var sheetname = sheet.name;
    var template =
        "<li class='sheetTab'>" +
        "<a href='#sheet_<%= name %>' class='sheetType_<%= type %>'>" +
        "<%= name %>" +
        // "<br> <span class='sheetType'>(<%= type %>)</span>" +
        "<span class='ui-icon ui-icon-close' style='display:inline-block;' role='presentation'></span>" +
        "</a>" +
        "</li>";
    li = _.template(template, sheet);
    var tabs = $("#sheets");

    tabs.find(".ui-tabs-nav").append(li);
    tabs.append(
        "<div id='sheet_" + sheetname + "' class='sheetContainer'>" +
        "    <canvas class='sheet' id='canvas_" + sheetname + "'></canvas>" +
        "</div>");
    tabs.tabs("refresh");
    setupSelectable(sheetname);
    // update network data structure.
    updateSheetsSize();
    clearCanvas("canvas_" + sheetname);

    addSheetScrollEvent(sheetname);

    tabs.tabs({active: getSheetId(sheetname)});
    $(".sheetTab[aria-controls='sheet_" + sheet.name + "']").tooltip(
        {position: {my: 'center bottom', at: 'center top-5'}});
    setTimeout(function(){sheetsView.toggleMinimize();}, 100);
}

function removeSheet(sheetname, noAnimation){
    var obj = $("a[href=#" + sheetname + "]").closest("li");
    if (noAnimation){
        obj.remove();
        $("#" + sheetname).unbind().remove();
    } else {
        obj.fadeOut(function(){obj.unbind().remove();});
        $("#" + sheetname).fadeOut(function(){$("#" + sheetname).unbind().remove();});
    }

    // from sheet_??? --> ???
    sheetname = sheetname.substring(sheetname.indexOf("_") + 1);
    delete network.sheets[sheetname];
    removeSheetScrollEvent(sheetname);

    //update network data structure.
    var removedNodes = _.filter(network.nodes,
        function(node){return node.sheet == sheetname;});

    // remove the link if its fromNodeId or toNodeId is removedNode.
    network.links = _.reject(network.links,
        function(link){
            return _.find(removedNodes,function(n){return n.id === link.fromNodeId || n === link.toNodeId;});});
    network.links = _.reject(network.links,
        function(l){return l.fromNodeId.indexOf(sheetname) !== -1 ||
                           l.toNodeId.indexOf(sheetname) !== -1;});

    // remove dynamic node and the nodes in the removed sheet.
    _.chain(network.nodes)
     .filter(function(n){return n.sheet === sheetname || n.name === sheetname;})
     .each(function(n){removeNode(n.id, false, noAnimation);});

    // remove from the nodedef
    $("#flatNodelist [key=" + sheetname + "]").unbind().remove();
    $(_.chain($("#nodelist [role=menuitem]"))
       .find(function(n){return $(n).text().trim() === sheetname;})
       .value()).parent().unbind().remove();

    $("#sheet").tabs("refresh");

    // drawLinks();
}

function setupSheets(){
    $("#sheets").tabs({
        activate: function(event, ui){
            currentTab = ui.newTab.text();
            _.chain(network.nodes)
             .filter(function(n){return n.sheet === currentTab;})
             .each(function(n){$("#" + n.id).offset({left: n.pos[0], top: n.pos[1]});});
            updateSheetsSize();
            clearCanvas();
            drawLinks();
        }
    });
    // $("#sheets .ui-tabs-nav").css("overflow-y", "hidden").css("overflow-x", "scroll");

    $("#addSheetButton").button({
        text: false,
        icons: {primary: "ui-icon-plusthick"}
    }).click(function(){dialog.dialog("open");});

    $("#reshapeButton").button({
        text: false,
        icons: {primary: "ui-icon-arrowrefresh-1-s"}
    }).click(function(){
        _.each(_.keys(network.sheets), function(sheetname){
            // var sheetname = $("#sheets .ui-tabs-active").text();
            $("#sheet_" + sheetname.name).scrollTop(0);
            $("#sheet_" + sheetname.name).scrollLeft(0);
        });
        var prevActiveTab = $("#sheets").tabs("option", "active");

        $.ajax({
            type: "PUT",
            url: rootPath + "reshapedNodes",
            dataType: "json",
            data: network
        }).done(function(reshaped, dataType){
            if (!reshaped.nodes){return;}
            network = reshaped;
            _.each(network.nodes, function(n){
                n.pos[0] += truncate($("#sheets").offset().left) + 60;
                n.pos[1] += truncate($("#sheets").offset().top ) + 90;
            });
            _.each(network.nodes, function(n){
                console.log(n.id, n.pos[0], n.pos[1]);
            });
            _.each(_.range(_.keys(reshaped.sheets).length), function(id){
                $("#sheet").tabs({active: id});
                var sheetname = $("#sheets .ui-tabs-active").text();
                $("#sheet_" + sheetname).scrollTop(0);
                $("#sheet_" + sheetname).scrollLeft(0);
                _.delay(function(){
                    _.chain(reshaped.nodes)
                     .filter(function(n){return n.sheet === sheetname;})
                     .each(function(n){
                        $("#" + n.id).offset({left: n.pos[0], top: n.pos[1]});
                     });
                    clearCanvas("canvas_" + sheetname);
                    drawLinks("canvas_" + sheetname);
                }, 300);
            });
            $("#sheets").tabs({active: prevActiveTab});
        }).fail(function(data){
            alert("you need to install Graphviz.");
            // $("#reshapeButton").tooltip();
            // $("#reshapeButton").tooltip("destroy");
            // $("#reshapeButton").remove();
        });
    });

    $("#AddIODialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            Add: function(){
                var clickedNodeId = $(this).attr("clickedNodeId");
                var clickedNodeObj = _.findObject(network.nodes, "id", clickedNodeId);
                var terminalRolesOfClickedNode =
                    _.chain(network.terminalRoles)
                     .filter(function(n){return n.nodeId === clickedNodeId;})
                     .value();
                var type = ($(this).attr("title").indexOf("Input") === -1) ? "outputs" : "inputs";
                var value = $("#IOName").val();

                if (_.pickArray(clickedNodeObj.inputs,  "name").indexOf(value) !== -1 ||
                   _.pickArray(clickedNodeObj.outputs, "name").indexOf(value) !== -1){
                    raise(clickedNodeId + " names are duplicated.", value);
                } else {
                    clickedNodeObj[type].push({
                        description: "Dynamically added " + type,
                        name: value,
                        type: "any"
                    });
                    type = (type == "inputs") ? "in" : "out";
                    $("#" + clickedNodeId + " ." + type + "terminals").html(
                        _.template($("#nodeIOTemplate").html(),
                                   {node: clickedNodeObj, type: type, term: terminalRolesOfClickedNode})
                    );
                }
                $(this).dialog("close");
            },
            Cancel: function(){$(this).dialog("close");}
        }
    });
    var dialog = $("#addSheetDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            Add: function(){
                var sheetname = $("#addSheetName").val();
                var sheettype = $("#addSheetType").val();
                addSheet(sheetname, sheettype);
                $("#addSheetName").val("");
                $(this).dialog("close");
            },
            Cancel: function(){$(this).dialog("close");}
        }
    });

    $("#addSheetDialog").find("form").submit(function(event){
        $($('#addSheetDialog').siblings('.ui-dialog-buttonpane').find("button")[0]).click();
        dialog.dialog("close");
        event.preventDefault();
        event.stopPropagation();
    });

    $("#AddIODialog").find("form").submit(function(event){
        addSheet($("#addSheetName").val(), $("#addSheetType").val());
        dialog.dialog("close");
        event.preventDefault();
        event.stopPropagation();
    });

    $("#sheets").on("click","span.ui-icon-close",  function(){
        var sheetname = $(this).closest("li").attr("aria-controls");
        if (confirm("Are you sure to remove " + sheetname + "?")){
            removeSheet(sheetname);
            $("#sheets").tabs({active: 0});
        }
    });
    $("#sheets").on("click", function(){
        $("#nodelistSearch input").blur();
    });
    setupSelectable("MAIN");
}
function setupSelectable(sheetname){
    $("#sheet_" + sheetname).selectable({
        filter: ".node",
        cancel: "#nodelist",
        distance: 30,
        selecting: function(e, ui){
            $(ui.selecting).addClass("nodeSelected");
            // $(".nodeSelected").draggable();
        },
        unselecting: function(e, ui){
            $(ui.unselecting).removeClass("nodeSelected");
        },
        start: function(e, ui){
            $(".nodeSelected").removeClass("nodeSelected");
        }
    });
    $("#sheet_" + sheetname).on("click", function(e){
        $(".nodeSelected").removeClass("nodeSelected");
    });
}

/* Load and parse .def file for nodelist*/
function loadNodeListAjax(defname){
    rootPath = rootPath || "";
    var tmpdef = defname.replace(".def", "");

    // load .def file
    $.ajax({
        type: "GET",
        url: rootPath + "def/" + defname,
        dataType: "json",
        beforeSend: function(){
            $(".nodeLoading").append(
                "<div class ='def def" + tmpdef + "'>" + defname + "...</div>");
            $(".nodeLoading").height(
                parseInt($(".nodeLoading").height(), 10) + 20);
        }
    }).done(function(data){
        _.each(data, parseDefFile);
        // parseDefFile(xml);
        setSearchEvents();
        userPreference.addPackage(defname);

        // update display for new nodelist
        updateNodeList();
        $("#sidebar").accordion("refresh");
        $("#sidebar").css("z-index", 1000);
        $("#sheets .ui-tabs-nav").css("height", "43px");
        updateSheetsSize();

        $(".nodeLoading .def" + tmpdef).text(defname + "... done.");
        // $(".nodeLoading").height(parseInt($(".nodeLoading").height()) - 18)
        updateNodeList();
    }).fail(function(data){
        alert("Reading " + defname + " failed.\n You can load from Preference -> Packages.");
        userPreference.delPackage(defname);
    });

    //xml parser
    function parseDefFile(def){
        var nodeobj = $(this);
        if (_.indexOf(_.pluck(nodedefs, "name"), def.name) !== -1){
            return;
        }

        nodedefs.push({
            name:        def.name,
            category:    def.category,
            description: def.description.replace(/[ \n\r]*$/g, ""),
            inputs:      [],
            outputs:     [],
            parameters:  [],
            def:         defname,
            theme:       "Classic"
        });
        _.each(def.inputs, function(i){
            _.last(nodedefs).inputs.push({
                name: i.name,
                type: i.object_type,
                description: i.description.replace(/[ \n\r]*$/g, "")
            });
        });
        _.each(def.outputs, function(o){
            _.last(nodedefs).outputs.push({
                name: o.name,
                type: o.object_type,
                description: o.description.replace(/[ \n\r]*$/g, "")
            });
        });
        _.each(def.parameters, function(p){
            var paramobj = $(this);
            var paramlist = [];
            newParameter(p, _.last(nodedefs));
        });
    }
}
function updateNodeList(){
    var nodelistdom = $("#nodelist");
    var flatnodelistdom = $("#flatNodelist");
    var nodedefsPerCategory = _.groupBy(nodedefs, "category");
    var c_tmpl = '<li type="category" key="<%= category %>" category="<%= category %>"><a href="#"><%= category %></a></li>';
    var n_tmpl = '<li key="<%= name %>" class="nodeElement" title="<%= name %>: <%= description %>"><a href="#"><%= name %></a></li>';
    var f_tmpl = '<li key="<%= name %>" category="<%= category %>" class="nodeElement" title="<%= description %>"><%= name %></li>';

    // existing nodes and categories
    var datCats = _.sortBy(_.keys(nodedefsPerCategory),
                           function(c){return c;});
    var domCats = _.sortBy(_.map(nodelistdom.children("li"), function(c){return $(c).attr("key");}),
                           function(c){return c;});
    var newCats = _.reject(datCats, function(c){return _.contains(domCats, c);});
    var extCats = _.reject(datCats, function(c){return _.contains(newCats, c);});
    var delCats = _.reject(domCats, function(c){return _.contains(datCats, c);});

    // delete removed categories
    _.each(delCats, function(c){
        nodelistdom.find("[key='" + c + "']").unbind().remove();
        flatnodelistdom.find("[key='" + c + "']").unbind().remove();
        flatnodelistdom.find("[category='" + c + "']").unbind().remove();
    });

    // add new category
    _.each(newCats, function(category){
        nodelistdom.append(_.template(c_tmpl, {category: category}));
        nodelistdom.children("li:last").append("<ul></ul>");
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}), function(n){
            nodelistdom.find("ul:last").append(_.template(n_tmpl, n));
        });

        flatnodelistdom.append(_.template(c_tmpl, {category: category}));
        flatnodelistdom.children("li:last").append("<ul></ul>");
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}), function(n){
            flatnodelistdom.find("ul:last").append(_.template(n_tmpl, n));
            flatnodelistdom.append(_.template(f_tmpl, n));
        });
    });

    // update existing category
    _.each(extCats, function(category){
        var catdom = nodelistdom.find("li[type=category][key='" + category + "'] ul");
        var datName = _.pluck(nodedefsPerCategory[category], "name");
        var domName = _.map(catdom.children("li"), function(l){return $(l).attr("key");});

        if (_.all(_.map(_.zip(datName.sort(), domName.sort()), function(pair){return pair[0] === pair[1];}))){
            return;
        }

        catdom.html("");
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}),
               function(n){catdom.append(_.template(n_tmpl, n));});

        catdom = flatnodelistdom.find("li[type=category][key='" + category + "'] ul");
        catdom.html("");
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}),
            function(n){catdom.append(_.template(n_tmpl, n));});

        flatnodelistdom.children(".nodeElement[category='" + category + "']").unbind().remove();
        _.each(_.sortBy(nodedefsPerCategory[category], function(n){return n;}),
            function(n){flatnodelistdom.append(_.template(f_tmpl, n));});
    });
    nodelistdom.menu("refresh");
    flatnodelistdom.menu("refresh");
    $("#flatnodelist .nodeElement").tooltip({position: {my: 'left center', at: 'right+5 center'}});
    $("#nodelist .nodeElement").tooltip({position: {my: 'left center', at: 'right+5 center'}});

    nodelistSize();
}
function nodelistSize(){
    $("#sidebar").css("width", "auto");
    $("#sheets").css("margin-left", $("#sidebar").width() + "px");
    $("#sidebar").css("width", $("#sidebar").width() + "px");
}
var isSearchEventsAdded = false;
function setSearchEvents(){
    if (isSearchEventsAdded){
        return;
    } else {isSearchEventsAdded = true};
    //search
    // when the Enter key is pressed
    $("#nodelistSearch").submit(function(event){
        var node = $("#flatNodelist li.active");
        if (node.length > 0){
            $(node).click();
        }
        $("#flatNodelist li").removeClass("active");
        event.preventDefault();
        event.stopPropagation();
        return false;
    });

    //incremental search
    var isPrevMin;
    $("#nodelistSearch input").keyup(function(event){
        var query = $(this).val().toLowerCase();
        if (event.keyCode === 13){
            userPreference.set("nodelistmin", false);
        }
        if ([13, 37, 38, 39, 40].indexOf(event.keyCode) !== -1){
            return;
        }

        if (!query && prevQuery === query){
            return;
        }

        $(".nodeElement").removeClass("active");
        $(".nodeSelected").removeClass("nodeSelected");

        var nodelist = [];
        if (query === ""){
            isSearching = false;
            $("#nodelist").show();
            $("#flatNodelist").hide();
            nodelistSize();
            userPreference.set("nodelistmin", isPrevMin);
            isPrevMin = undefined;
        } else {
            if (isPrevMin === undefined){
                isPrevMin = userPreference.get("nodelistmin");
            }
            userPreference.set("nodelistmin", false);
            $("#nodelist").hide();
            $("#flatNodelist").show();
            //search

            $("#flatNodelist > li").each(function(){
                var sttag = '<span style="font-weight:bold;">';
                var edtag = "</span><!--searching-->";
                var html = $(this).html();

                var pos = $(this).attr("key").toLowerCase().indexOf(query);
                if (pos > -1){
                    if (html.indexOf("</ul>") === -1){
                        html = html.replace(sttag, "").replace(edtag, "");
                        var st = pos, ed = pos + query.length;
                        html = html.substring(0, st) + sttag +
                               html.substring(st, ed) + edtag +
                               html.substring(ed);
                        $(this).html(html);
                    }
                    $(this).show();
                } else {
                    $(this).hide();
                }
            });
            if ($("#flatNodelist > li:visible").length === 0){
                $("#nodeNotFound").show();
            } else {
                $("#nodeNotFound").hide();
            }
        }
        nodelistSize();
        prevQuery = query;
    });
    $(document).on("mouseover", ".nodeElement", function(){$(".nodeElement").removeClass("active"); $(this).addClass("active");});
    $(document).on("mouseout",  ".nodeElement", function(){$(this).removeClass("active");});
}

function data2graphviz(){
    var out = "";
    out += "digraph network{\n";
    _.chain(network.links)
     .each(function(link){
        out += link.fromNodeId + "->" + link.toNodeId + ";\n";
    });
    out += "}\n";
}

/**** network manipulation functions with verification ****/
function raise(message, obj){
    if (!message){
        app.set("execstat", "");
        return;
    }
    console.log("Error: " + message + " (" + JSON.stringify(obj) + ")");
    app.set("execstat", "Error: " + message);
}

function hasMandatory(obj, mandatory){
    if (_.chain(mandatory)
        .map(function(m){return _.has(obj, m);})
        .every().value() === false){
        return false;
    }
    return true;
}

function isTypesMatch(type1, type2){
    if (type1 === "any" || type2 === "any"){
        return true;
    }
    return type1 === type2;
}

function newNode(node, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var result = testNode(node, tmpnetwork);
    raise(result, node);
    if (result !== ""){
        return false;
    }
    tmpnetwork.nodes.push(node);
    return true;
}

function newLink(link, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var result = testLink(link, tmpnetwork);
    raise(result, link);
    if (result !== ""){
        return false;
    }
    tmpnetwork.links = tmpnetwork.links || [];
    tmpnetwork.links.push(link);
    return true;
}

function newTerminalRole(role, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var result = testTerminalRole(role, tmpnetwork);
    raise(result, role);
    if (result !== ""){
        return false;
    }
    // console.log(tmpnetwork);
    tmpnetwork.terminalRoles = tmpnetwork.terminalRoles || [];
    tmpnetwork.terminalRoles.push(role);
    updateTerminalRoles(role, "add");
    return true;
}

function newSheet(sheet, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var result = testSheet(sheet, tmpnetwork);
    raise(result, sheet);
    if (result !== ""){
        return false;
    }
    tmpnetwork.sheets[sheet.name] = sheet;
    return true;
}

function newParameter(param, node){

    // prune parameter types
    var mapper = {
        any: "int",
        String: "string"
    };
    if (mapper[param.type]) {param.type = mapper[param.type];}
    if (param.type.indexOf("Vector") !== -1) {param.type = "object";}

    var result = testParameter(param);
    raise(result, param);
    if (result !== ""){
        return false;
    }

    // new parameter name cannot be duplicated.
    if (_.indexOf(_.pickArray(node.parameters, "name"), param.name) !== -1){
        raise("Parameter name cannot be duplicated", param);
        return false;
    }
    node.parameters.push(param);
    return;
}

function testParameter(param){
    // condition: if not empty, the format must be *=*.
    if (!_.isEmpty(param.condition) && param.condition.indexOf("=") === -1 ){
        return param.name + " Parameter coondition must include =";
    }

    // description: no empty
    // if (_.isEmpty(param.description)){
    //     return param.name + " Parameter description cannot be empty";
    // }

    // list       : must be list
    if (!_.isArray(param.list)){
        param.list = [];
        // return param.name + " Parameter list must be array: " + param.list;
    }

    // name       : not empty
    if (_.isEmpty(param.name)){
        return param.name + " Parameter name cannot be empty.";
    }

    // type       : [int, string, bool, object, float, any, String, subnet_param]
    var types = ["int", "string", "bool", "object", "float", "any", "String", "subnet_param", "file"];
    if (_.indexOf(types, param.type) === -1){
        return param.name + " Parameter type unknown (" + param.type + ")";
    }

    // value      : depends on the type
    if (!_.isEmpty(param.value)){
        switch (param.type){
            case "int":
            if (_.isString(param.value) && param.value.match(/^-?\d+$/) === null){
                return param.name + " Parameter format is not integer.";
            }
            param.value = parseInt(param.value, 10);
            if (!_.isNumber(param.value)){return param.name + " Parameter value is not number";}
            if (_.isNaN(param.value)){return param.name + " Parameter value is NaN (int)";}
            if (parseFloat(param.value) % 1 !== 0){return param.name + " Parameter value is not integer";}
            break;

            case "float":
            if (_.isString(param.value) && param.value.match(/^-?[\d.]+$/) === null){
                return param.name + " Parameter format is not float.";
            }
            param.value = parseFloat(param.value);
            if (!_.isNumber(param.value)){return param.name + " Parameter value is not number";}
            if (_.isNaN(param.value)){return param.name + " Parameter value is NaN (float)";}
            break;

            case "object":
            // < and > must match
            if (param.value.split("<").length != param.value.split(">").length){
                return param.name + " Parameter value parenthesis <> mismatched";
            }
            if (param.value.indexOf(">>") != -1){
                return param.name + " Parameter value >> must be > >.";
            }
            break;
        }
    }
    return "";
}

function testNode(node, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var mandatory = ["name", "id", "sheet", "index", "pos", "parameters", "inputs", "outputs"];
    if (!hasMandatory(node, mandatory)){
        return node.id + " node does not have mandatory options";
    }

    // name: defined,
    if (_.findObject(nodedefs, "name", node.name) === undefined){
        return node.id + " unknown node name";
    }

    // id  : node_<name>_<index> format
    if (node.id !== "node_" + node.name + "_" + node.index){
        return node.id + " wrong node id format";
    }

    // sheet: in network.sheets, not(node.name === node.sheet)
    if (_.keys(tmpnetwork.sheets).indexOf(node.sheet) === -1){
        return node.id + " node sheet is undefined";
    }
    if (node.name === node.sheet){
        return node.id + " The dynamic node cannot be placed in the same sheet";
    }

    // index: integer, positive, unique
    if (!_.isNumber(node.index)){
        return node.id + " node index must be integer";
    }
    if (node.index <= 0){
        return node.id + " node index must be positive";
    }
    if (_.chain(tmpnetwork.nodes)
        .filter(function(n){return n.name === node.name;})
        .map(function(n){return n.index;}).value().indexOf(node.index) !== -1){
        return node.id + " Node index is duplicated";
    }

    // pos  : positive, two-length array
    // if (!(_.isArray(node.pos) && node.pos.length === 2)){
    //     return node.id + " Node pos must be an array with length = 2";
    // }
    // if (node.pos[0] < 0){return node.id + " node position[0] is not positive";}
    // if (node.pos[1] < 0){return node.id + " node position[1] is not positive";}

    // parameters: array
    if (!_.isArray(node.parameters)){
        return node.id + " Node parameters must be an array!";
    }

    return "";
}

function escapeHTML(text){
    return $("<div>").text(text).html();
}

function testLink(link, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }

    var mandatory = ["fromNodeId", "fromTermName", "toNodeId", "toTermName", "sheet"];
    if (!hasMandatory(link, mandatory)){
        return "Link does not have mandatory properties";
    }

    var fromNodeObj = _.findObject(tmpnetwork.nodes, "id", link.fromNodeId);
    var toNodeObj = _.findObject(tmpnetwork.nodes, "id", link.toNodeId);

    // fromNodeId: node_*_\d, in nodes
    if (link.fromNodeId.match(/node_.+_\d/) === null){
        return "Link fromNodeId (" + link.fromNodeId + ") is invalid format";
    }
    if (fromNodeObj === undefined){
        return "Link fromNodeId (" + link.fromNodeId + ") is not found in nodes";
    }

    // toNodeId: node_*_\d, in nodes
    if (link.toNodeId.match(/node_.+_\d/) === null){
        return "Link toNodeId (" + link.toNodeId + ")is invalid format";
    }
    if (toNodeObj === undefined){
        return "Link toNodeId (" + link.toNodeId + ") is not found in nodes";
    }

    var fromDef = _.findObject(nodedefs, "name", fromNodeObj.name);
    var toDef = _.findObject(nodedefs, "name", toNodeObj.name);
    var fromTerms = _.pickArray(fromNodeObj.outputs, "name");
    var toTerms = _.pickArray(toNodeObj.inputs, "name");

    // fromTermName: in nodedef.outputs
    // if (fromTerms.indexOf(link.fromTermName) === -1){
    //     return "Link fromTermName is not defined in outputs";
    // }

    // toTermName: in nodedef.inputs
    // if (toTerms.indexOf(link.toTermName) === -1 &&
    //    toDef.category !== "Dynamic"){
    //     return "Link toTermName is not defined in inputs";
    // }

    // matching: *nodeId cannot be the same.
    if (link.fromNodeId === link.toNodeId){
        return "Link cannot connect the same nodes (" + link.toNodeId + ")";
    }

    // sheetname: fromNodeObj.sheet === toNodeObj.sheet === sheets
    if (fromNodeObj.sheet !== toNodeObj.sheet){
        return "Link cannot connect different sheets";
    }
    if (fromNodeObj.sheet !== link.sheet){
        return "Link node.sheets and link.sheets are different";
    }

    // matching: object_type must be matched.
    var fromType = _.findObject(fromNodeObj.outputs, "name", link.fromTermName);
    var toType   = _.findObject(toNodeObj.inputs, "name", link.toTermName);
    if (!fromType || !toType){
        return "Link cannot connect INPUT-INPUT nor OUTPUT-OUTPUT";
    }

    if ( toDef.category !== "Dynamic" && fromDef.category !== "Dynamic" &&
        !isTypesMatch(fromType.type, toType.type)){
        return "Link input and output types are mismatched " +
                "from type is " + escapeHTML(fromType.type) + ",     but " +
                "to type is " + escapeHTML(toType.type);
    }

    // toTerms cannot have multiple inputs
    var wrongTo = _.find(tmpnetwork.links, function(l){
        return l.toNodeId === link.toNodeId && l.toTermName === link.toTermName;
    });
    if (wrongTo){
        return "Link input terminal (" + wrongTo.toTermName + " of " + wrongTo.toNodeId + ") cannot take multiple links";
    }

    return "";
}

function testTerminalRole(role, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var mandatory = ["nodeId", "termName", "name", "role", "sheet"];
    if (!hasMandatory(role, mandatory)){
        return "role does not have mandatory properties";
    }
    if (role.role === "input" || role.role === "output"){
        if (!hasMandatory(role, ["object_type", "description"])){
            return "role does not have mandatory properties";
        }
    }

    // nodeId: format, in tmpnetwork.nodes
    if (role.nodeId.match(/node_.+_\d/) === null){
        return "Role nodeId has wrong format";
    }
    var thenode = _.findObject(tmpnetwork.nodes, "id", role.nodeId);
    if (thenode === undefined){
        return "Role nodeId does not exist in the network";
    }

    // role: input/output/condition
    if (["input", "output", "condition"].indexOf(role.role) === -1){
        return "Role role must be input / output / condition";
    }

    // termName: input -> in node.input, output/condition -> in node.output
    // var def = _.find(nodedefs, function(d){return d.name === thenode.name});
    var def = _.findObject(nodedefs, "name", thenode.name) | {
        category: "Undefined"
    };

    if (role.role === "input"){
        if (def.category !== "Dynamic" &&
           _.findObject(thenode.inputs, "name", role.termName) === undefined){
            return "Role termname does not found in inputs";
        }
    } else {
        if (def.category !== "Dynamic" &&
           _.findObject(thenode.outputs, "name", role.termName) === undefined){
            return "Role termname does not found in outputs";
        }
    }

    // sheet in network.sheets, nodeId.sheet === sheet
    if (_.keys(tmpnetwork.sheets).indexOf(role.sheet) === -1){
        return "Role sheet does not exist";
    }

    if (role.sheet !== thenode.sheet){
        return "Role sheet and nodeId.sheet mismatched";
    }

    // name: not empty, no conflict in the same sheet, not empty
    if (_.isEmpty(role.name)){
        return "Role name cannot be empty";
    }

    var existingNames = _.pickArray(_.groupBy(tmpnetwork.terminalRoles, "sheet")[role.sheet], "name");
    if (existingNames.indexOf(role.name) !== -1){
        return "Role name already exist in the sheet";
    }

    // object_type: if not condition --> not empty
    // if (role.role !== "condition" && _.isEmpty(role.object_type)){
    //     return "Role object_type cannot be empty";
    // }
    // description: no check

    return "";

// nodeId   : nodeId,
// termname : termname,
// name     : rolename,
// role     : "input",
// sheet    : sheetname,
// object_type: objtype,
// description: desc
}

function testSheet(sheet, tmpnetwork){
    if (!tmpnetwork){
        tmpnetwork = network;
    }
    var mandatory = ["name", "type"];
    if (!hasMandatory(sheet, mandatory)){
        return "Sheet does not have mandatory property";
    }

    // type cannot be empty
    if (_.isEmpty(sheet.type)){
        return "Sheet type cannot be empty";
    }

    // type must be the list
    var typelist = ["subnet", "iterator"];
    if (typelist.indexOf(sheet.type) === -1){
        return "Sheet type must be subnet or iterator";
    }

    // no duplicate
    if (_.keys(tmpnetwork.sheets).indexOf(sheet.name) !== -1){
        if (sheet.name !== "MAIN"){
            return "Sheet name is duplicated";
        }
    }

    // not empty
    if (_.isEmpty(sheet.name)){
        return "Sheet name cannot be empty";
    }

    return "";
}

function pruneNetwork(tmpnetwork){
    // var nodeIds = _.pickArray(network.nodes, "id");
    // network.links = _.chain(network.links)
    // .filter(function(l){
    //     return nodeIds.indexOf(l.fromNodeId) !== -1 &&
    //            nodeIds.indexOf(l.toNodeId);
    // })
    // .filter(function(l){
    //     var node = _.findObject(network.nodes, "id", l.fromNodeId);
    //     return _.findObject(node.outputs, "name", l.fromTermName);
    // })
    // .filter(function(l){
    //     var node = _.findObject(network.nodes, "id", l.toNodeId);
    //     return _.findObject(node.inputs, "name", l.toTermName);
    // }).value();

    _.each(tmpnetwork.nodes, function(n){
        n.pos[0] = truncate(n.pos[0]);
        n.pos[1] = truncate(n.pos[1]);
    });

    if (typeof(tmpnetwork.links) === "undefined"){
        tmpnetwork.links = [];
    }
    if (typeof(tmpnetwork.terminalRoles) === "undefined"){
        tmpnetwork.terminalRoles = [];
    }
}

function setupSaveFile(){
    $("#saveFileDialog").submit(function(event){
        if (app){
            $($('#saveFileDialog').siblings('.ui-dialog-buttonpane').find("button")[0]).click();
        }
        event.preventDefault();
        event.stopPropagation();
    });
    $("#saveFileDialog").dialog({
        autoOpen: false,
        modal: true,
        buttons : {
            OK: function(){
                var that = this;
                var filename = $("#saveFileDialog input[name=name]").val();

                if (_.isEmpty(filename)){
                    alert("You must specify the file name");
                    return;
                }
                if (_.pluck(fileCollection.toJSON(), "name").indexOf(filename) !== -1 &&
                   app.get("currentFileName") !== filename){
                    if (!confirm("Overwrite network?")) {return;}
                }
                app.set("currentFileName", filename);

                // replace all "file" type parameters to "string"
                _.each(network.nodes, function(n){
                    _.each(n.parameters, function(p){
                        if (p.type === "file"){
                            p.type = "string";
                            // console.log(p);
                        }
                    });
                });
                // console.log(JSON.stringify({object: network}));

                $.ajax({
                    url: rootPath + "file/object/" + filename,
                    type: "PUT",
                    data: network,
                    // dataType: "json"
                }).done(function(){
                    app.set("execstat", "Saved");
                    updateFileList();
                    $(that).dialog("close");
                }).fail(function(data){
                    console.log("failed", data);
                    app.set("execstat", data.responseText);
                });
            },
            Cancel: function(){$(this).dialog("close");}
        }
    });
    $("#save")
        .click(function(event){
            $("#saveFileDialog").dialog("open");
        });
    updateFileList();
}

/* update canvas size */
function updateSheetsSize(){
    _.each($("#sheets canvas"), function(canvas){
        canvas = $(canvas);
        canvas.attr("height", canvas.css("height"));
        canvas.attr("width", canvas.css("width"));
    });

    // one implementation: canvas size = sidebar size
    // var side = parseInt($("#sidebar").css("height"), 10);
    // _.each($("#sheets .sheetContainer"), function(div){
    //     $(div).css("height", (side - 50) + "px");
    // });
    // $("#sheets").css("height", side + "px");
    // return;

    // another implementation: canvas size = window size
    var body = $("body").height();
    var footer = $("#footer").height();
    var padding = 30;
    // console.log(body, footer, padding);
    _.each($("#sheets .sheetContainer"), function(div){
        $(div).css("height", (body - $(div).offset().top - footer - padding) + "px");
    });

    $("#nodelistContainer").css("height",
        (body - $("#nodelistContainer").offset().top - footer - padding) + "px");
}
/* setup save network to file list */
function updateFileList(){
    fileCollection.reset();
    fileCollection.fetch({url: rootPath + "filelist", add: true});
}

function setupPreference(){
    $("#preferenceButton").click(function(event){
        $("#preferenceDialog").dialog("open");
    });

    $("#preferenceDialog").dialog({
        autoOpen: false,
        modal: true,
        minWidth: 700,
        open: function(){
            _.each($("#preferenceDialog .dialogMenu button"), function(li){
                $(li).css("width", "100%");
            });

            $("#preferenceDialog .dialogMain > div").css("display", "none");
        },
        close: function(){
            $("#preferenceDialog .dialogMenu button").removeClass("ui-state-highlight");
            $("#preferenceDialog .dialogMain > div").css("display", "none");
        }
    });

    $("#preferenceDialog .package").on("change", "form", function(event){
        var newTheme = event.target.value;
        var selectedDef = $(event.target).siblings("label").text();

        _.chain(nodedefs)
         .filter(function(d){return d.def === selectedDef;})
         .each(function(d){d.theme = newTheme;});

         _.chain(network.nodes)
          .filter(function(n){return n.def === selectedDef;})
          .each(function(n){
            $("#" + n.id).removeClass("nodeTheme" + n.theme);
            $("#" + n.id).addClass("nodeTheme" + newTheme);
            n.theme = newTheme;
          });
    });
    var prefEnvironment = function(data){
        function addVarForm(that, name, variable){
            var num = $("#preferenceDialog .dialogMain .setenv fieldset").length || 1;
            if (!_.isNumber(num)){
                num = _.max(_.map(num, function(n){
                    console.log($(n).attr("id").replace("variable", ""), 10);
                    return parseInt($(n).attr("id").replace("variable", ""), 10) + 1;}));
            }

            var html = _.template(
                '<fieldset id="variable<%= num %>" class="val">' +
                '    <input id="varname<%= num %>" <% if (name) {%> value="<%= name %>" <% } %>>' +
                '    <input id="varval<%= num %>"  <% if (variable) {%> value="<%= variable %>" <% } %>>' +
                '    <button id="delvar<%= num %>">' +
                '</fieldset>',
                {num: num, name: name, variable: variable});
            $(that).siblings(".setenv").append(html);
            $("#preferenceDialog .dialogMain #delvar" + num).button({
                text: false,
                icons: {primary: "ui-icon-minusthick"}
            }).click(function(event){
                $(this).parent().unbind().remove();
            });
        }

        $("#preferenceDialog .dialogMain .package").html(data);
        $("#preferenceDialog .dialogMain .package").css("display", "block");
        $("#preferenceDialog .dialogMain .addvar").button({
            label: "Add new variable",
            icons: {primary: "ui-icon-plusthick"}
        }).click(function(event){
            addVarForm(this);
        });

        _.each(_.keys(userPreference.get("env") || {}), function(e){
            if (e === "rosVersionInternal") {return;}
            addVarForm($("#preferenceDialog .dialogMain .addvar"),
                       e, userPreference.get("env")[e]);
        });

        $("#preferenceDialog .dialogMain .setvariables").button().click(
            function (event){
                var variables = {};
                variables.rosVersionInternal = $(this).siblings(".setros").find("select").val();

                _.each($("#preferenceDialog .dialogMain .setenv fieldset.val"), function(f){
                    var inputs = $(f).children("input");
                    variables[$(inputs[0]).val()] = $(inputs[1]).val();
                });
                console.log(variables);
                userPreference.set("env", variables);
                $.ajax({
                    url: "/preference/environment",
                    method: "PUT",
                    data: variables
                }).always(function(data){
                    $(this).siblings(".msg", data);
                });
            }
        );

        $("#preferenceDialog .dialogMain .mypkg .file").fileupload({
            // dataType: "json",
            prependFiles: true,
            done: function(e){
                console.log("mypkg done");
                updateMyPkgList();
            },
            add: function(e, data){
                data.url = "/package/" + data.files[0].name;
                data.submit();
            }
        });

        // load localStorage
        if (userPreference.get("env")){
            var variables = userPreference.get("env");
            if (variables.rosVersionInternal){
                $("#preferenceDialog .dialogMain .setros select").val(variables.rosVersionInternal);
            }
            // delete variables["rosVersionInternal"];
            // var varnames = _.keys(variables);
            // for (var i = 0; i < varnames.length; i++){
                // $("#preferenceDialog .dialogMain .addvar").trigger("click");
                // var ins = $("#preferenceDialog .dialogMain fieldset:last");
                // $($(ins).find("input")[0]).val(varnames[i]);
                // $($(ins).find("input")[1]).val(variables[varnames[i]]);
            // }
        }
        updateMyPkgList();
    };

    var updateMyPkgList = function(e, data){
        $.ajax({
            url: "/package",
            method: "GET"
        }).done(function(data){
            $("#preferenceDialog .dialogMain .mypkglist button").button("destroy");
            $("#preferenceDialog .dialogMain .mypkglist").html("");
            var html = "<tr><th>Package name</th><th>.tlb</th><th>.def</th><th></th></tr>";
            _.each(_.keys(data), function(pkg){
                html += "<tr><td>" + pkg + "</td>";
                if (_.indexOf(data[pkg], "tlb") !== -1){
                    html += "<td>OK</td>";
                } else {
                    html += "<td>Missing</td>";
                }

                if (_.indexOf(data[pkg], "def") !== -1) {
                    html += "<td>OK</td>";
                } else {
                    html += "<td>Missing</td>";
                }
                html += "<td><button></button></td>";
                html += "</tr>";
             });
            $("#preferenceDialog .dialogMain .mypkglist").append(html);
            $("#preferenceDialog .dialogMain .mypkglist button").button({
                text: false,
                icons: {primary: "ui-icon-closethick"}
            }).click(function(e){
                var pkg = $($($(this).parentsUntil("tr").parent()).children("td")[0]).text();
                userPreference.delPackage(pkg + ".def");
                console.log(pkg);
                $.ajax({
                    url: "/package/" + pkg,
                    method: "DELETE"
                }).always(function(data){
                    console.log("done");
                    updateMyPkgList();
                    updateNodeList();
                });
            });
        });
    };

    var prefPackage = function(data){
        $("#preferenceDialog .dialogMain .package").html(data);
        $("#preferenceDialog .dialogMain .package").css("display", "block");

        var maxw = _.chain($("#preferenceDialog .dialogMain .package button"))
                    .map(function(b){return $(b).next().width();})
                    .max().value();
        _.chain($("#preferenceDialog .dialogMain .package button"))
         .each(function(b){
            $(b).next().css("width", maxw * 1.1)
                       .css("margin-left", "10px");
        });

        $("#preferenceDialog").dialog("option", "width", maxw * 1.1);
        _.each($("#preferenceDialog .dialogMain .package button"),
            function(b){
                var def = $(b).next().text();
                var icon = "ui-icon-bullet";
                if (userPreference.get("packages").indexOf(def) === -1){
                    icon = "ui-icon-radio-off";
                    $(b).siblings("select").attr("disabled", "disabled");
                    $(b).siblings("label").css("color", "#AAA");
                }

                var tmpdef = _.find(nodedefs, function(d){return d.def === def;});
                if (tmpdef){
                    $(b).siblings("select").val(tmpdef.theme);
                }

                $(b).button({
                    text: false,
                    icons: {primary: icon}
                }).click(function(event){
                    var isExist = userPreference.get("packages").indexOf(def) !== -1;
                    if (isExist){
                        // nodedefs = _.reject(nodedefs, function(d){return d.def === def;});
                        // userPreference.set("packages",
                        //     _.without(userPreference.get("packages"), def));
                        userPreference.delPackage(def);
                        $(this).children("span").removeClass("ui-icon-bullet");
                        $(this).children("span").addClass("ui-icon-radio-off");
                        $(b).siblings("select").attr("disabled", "disabled");
                        $(b).siblings("label").css("color", "#AAA");
                    } else {
                        if (userPreference.get("packages").indexOf(def) === -1){
                            // console.log("loading");
                            loadNodeListAjax(def);
                            userPreference.addPackage(def);
                            $(this).children("span").addClass("ui-icon-bullet");
                            $(this).children("span").removeClass("ui-icon-radio-off");
                            $(b).siblings("select").removeAttr("disabled");
                            $(b).siblings("label").css("color", "#000");
                        }
                    }
                    updateNodeList();
                    event.stopPropagation();
                    event.preventDefault();
                });
            }
        );
    };

    var prefVersion = function(data){
        console.log(data);
        console.log(data.replace(/\n/g, "<br>"));
        $("#preferenceDialog .dialogMain .version").html(data.replace(/\n/g, "<br>"));
        $("#preferenceDialog .dialogMain .version").css("display", "block");
    };

    $("#preferenceDialog .dialogMain .display select[name=line]").on("change", function(event){
        console.log(event.target.value);
        userPreference.set("link", event.target.value);
        clearCanvas();
        drawLinks();
    });
    $("#preferenceDialog .dialogMain .display select[name=lang]").on("change", function(event){
        console.log(event.target.value);
        userPreference.set("lang", event.target.value);
    });
    $("#preferenceDialog .dialogMain .display select[name=tooltip]").on("change", function(event){
        console.log(event.target.value);
        userPreference.set("showTooltip", event.target.value);
        if (event.target.value === "Yes"){
            $(document).tooltip("enable");
        } else {
            console.log("destroyed");
            $(document).tooltip("disable");
        }
    });
    $("#preferenceDialog .dialogMenu button").each(function(li){
        $(this).button().click(function(event){
            $("#preferenceDialog .dialogMenu button")
                .removeClass("ui-state-highlight");
            $(this).addClass("ui-state-highlight");

            $("#preferenceDialog .dialogMain > div").css("display", "none");
            var cmd = $(this).text().toLowerCase();
            if (cmd === "packages"){
                $.ajax({url: rootPath + "preference/" + cmd}).done(prefPackage);
            } else if (cmd === "version"){
                $.ajax({url: rootPath + "preference/" + cmd}).done(prefVersion);
            } else if (cmd === "display"){
                $("#preferenceDialog .dialogMain .display").css("display", "block");
            } else if (cmd === "all reset"){
                if (confirm("Do you really want to reset all preferences?")){
                    localStorage.clear();
                    window.location.reload();
                    window.onbeforeunload = undefined;
                }
            } else if (cmd === "environment"){
                console.log(cmd);
                $.ajax({url: rootPath + "preference/" + cmd}).done(prefEnvironment);
            }

        });
    });
    $("#tutorialButton").button().click(function(){
        $("#preferenceDialog").dialog("close");
        introJs().start();
    });
}

function setupFileManager(){
    $("#fileManagerButton").click(function(event){
        updateFileList();
        $("#fileManagerDialog").dialog("open");
    });
}

function setupButtons(){
    $("#clear").click(function(event){
        if (confirm("Are you sure to clear current network?")){
            reset();
        }
    });

    $("#helpButton").click(function(event){
        window.open("http://winnie.kuis.kyoto-u.ac.jp/HARK/document/hark-designer-" + userPreference.get("lang"));
    });
}

function setupStatus(){
    $("#clearExecuteStatus").button().click(function(e){
        $("#executeStatus").text("");
        $(this).css("visibility", "hidden");
    });
    $("#clearExecuteStatus .ui-button-text").css("padding", "2px");
    $("#clearExecuteStatus .ui-button-text").css("font-size", "15px");
    $("#clearExecuteStatus").css("visibility", "hidden");
    $("#clearExecuteStatus").css("float", "right");

    $("#executeStatus").parent().  on("mouseover", function(event){
        if ($("#executeStatus").text()){
            $("#clearExecuteStatus").css("visibility", "visible");
        }
    });
}

function setupNodeKeyBindings(){
    $(document).keydown(function(event){
        // skip if any dialog is opened
        var opened = _.filter($(".ui-dialog"), function(d){return $(d).css("display") !== "none";}).length;
        if (opened > 0){
            return true;
        }

        // $(".ui-state-default:focus").blur();
        var doNothing = false;

        var key = event.keyCode;
        // $(".nodeSelected").removeClass("nodeSelected");
        selectedNodeObjs = _.map($(".nodeSelected"), function(n){
            id = $(n).children(".nodeid").text().trim();
            return _.findObject(network.nodes, "id", id);
        });
        if ([37, 9, 38, 39, 40, 27, 46].indexOf(key) !== -1){
            event.preventDefault();
            event.stopPropagation();
        }

        if (selectedNodeObjs.length > 0){
            var selectedLinks;
            switch (key){
                case 67: //Ctrl + C
                if (event.ctrlKey){
                    selectedLinks = _.filter(network.links, function(l){
                        var nodeIds = _.pluck(selectedNodeObjs, "id");
                        console.log(nodeIds);
                        return nodeIds.indexOf(l.fromNodeId) !== -1 && nodeIds.indexOf(l.toNodeId) !== -1;
                    });
                    clipboard = {nodes: selectedNodeObjs, links: selectedLinks};
                    clipboard = JSON.parse(JSON.stringify(clipboard));
                }
                break;

                case 88: //Ctrl + X
                // alert("Currently, Cut is disabled. Use Ctrl + C instead");
                if (event.ctrlKey){
                    selectedLinks = _.filter(network.links, function(l){
                        var nodeIds = _.pluck(selectedNodeObjs, "id");
                        console.log(nodeIds);
                        return nodeIds.indexOf(l.fromNodeId) !== -1 && nodeIds.indexOf(l.toNodeId) !== -1;
                    });
                    clipboard = {nodes: selectedNodeObjs, links: selectedLinks};
                    clipboard = JSON.parse(JSON.stringify(clipboard));
                    _.each(selectedNodeObjs, function(selectedNodeObj){
                        removeNode(selectedNodeObj.id);
                    });
                }
                break;

                case 9: //Tab
                _.each(selectedNodeObjs, function(selectedNodeObj){
                    $(".nodeSelected").removeClass("nodeSelected");
                    var shownNodes = _.filter(network.nodes, function(n){return n.sheet === currentTab;});
                    var tmp = _.chain(shownNodes)
                               .pluck("id").indexOf(selectedNodeObj.id).value();
                    console.log(tmp + "--" + shownNodes.length);
                    if (tmp == shownNodes.length - 1){
                        selectedNodeObj = undefined;
                        $("#nodelistSearch input").focus();
                        console.log($("#nodelistSearch input"));
                    } else {
                        selectedNodeObj = shownNodes[tmp + 1];
                        $("#" + selectedNodeObj.id).addClass("nodeSelected");
                    }
                    adjustScroll(currentTab, function(n){return $("#" + n.id).hasClass("nodeSelected");});
                });
                break;

                case 37: //left
                case 72: // h
                _.each(selectedNodeObjs, function(selectedNodeObj){
                    selectedNodeObj.pos[0] -= 10;
                    $("#" + selectedNodeObj.id).offset(
                        {left: selectedNodeObj.pos[0], top: selectedNodeObj.pos[1]});
                });
                break;

                case 38: //up
                case 75: //k
                _.each(selectedNodeObjs, function(selectedNodeObj){
                    selectedNodeObj.pos[1] -= 10;
                    $("#" + selectedNodeObj.id).offset(
                        {left: selectedNodeObj.pos[0], top: selectedNodeObj.pos[1]});
                });
                break;

                case 39: //right
                case 76: //l
                _.each(selectedNodeObjs, function(selectedNodeObj){
                    selectedNodeObj.pos[0] += 10;
                    $("#" + selectedNodeObj.id).offset(
                        {left: selectedNodeObj.pos[0], top: selectedNodeObj.pos[1]});
                });
                break;

                case 40: //down
                case 74: //j
                _.each(selectedNodeObjs, function(selectedNodeObj){
                    selectedNodeObj.pos[1] += 10;
                    $("#" + selectedNodeObj.id).offset(
                        {left: selectedNodeObj.pos[0], top: selectedNodeObj.pos[1]});
                });
                break;

                case 27: //escape
                _.each(selectedNodeObjs, function(selectedNodeObj){
                    $("#" + selectedNodeObj.id).removeClass("nodeSelected");
                    selectedNodeObj = undefined;
                });
                break;

                case 46: //delete
                _.each(selectedNodeObjs, function(selectedNodeObj){
                    removeNode(selectedNodeObj.id);
                });
                break;

                case 13:
                var counter = 0;
                _.each(selectedNodeObjs, function(selectedNodeObj){
                    $("#" + selectedNodeObj.id + delim + "parameters")
                        .delay(counter * 1000).dialog("open");
                    counter ++;
                });
                break;

                default:
                doNothing = true;
                break;
            }
            if (!doNothing){
                clearCanvas();
                drawLinks();
            }
        } else if ($("#nodelistSearch input:focus") && !selectedNodeObj){
            selectedNodeObj = undefined;
            $(".nodeSelected").removeClass("nodeSelected");

            var selected = $("#flatNodelist li.active:visible");
            var visibleList = $("#flatNodelist li:visible");
            var selectedIndex;
            for (var i = 0; i < visibleList.length; i++){
                if ($(visibleList[i]).text() === selected.text()){
                    selectedIndex = i;
                }
            }
            switch (key){
                case 9: // tab
                if (network.nodes.length > 0){
                    $("#nodelistSearch input").blur();

                    var shownNodes = _.filter(network.nodes, function(n){return n.sheet === currentTab;});

                    selectedNodeObj = shownNodes[0];
                    $("#" + selectedNodeObj.id).addClass("nodeSelected");
                    adjustScroll(currentTab, function(n){return $("#" + n.id).hasClass("nodeSelected");});
                }
                break;

                case 38: //up
                if (selectedIndex === undefined){
                    $("#flatNodelist li:visible").last().addClass("active");
                } else {
                    $("#flatNodelist li").removeClass("active");
                    if (selectedIndex === 0){
                        $(visibleList[visibleList.length - 1]).addClass("active");
                    } else {
                        $(visibleList[selectedIndex - 1]).addClass("active");
                    }
                }
                break;

                case 40: //down
                if (selectedIndex === undefined){
                    $("#flatNodelist li:visible").first().addClass("active");
                } else {
                    $("#flatNodelist li").removeClass("active");
                    if (selectedIndex === visibleList.length - 1){
                        $(visibleList[0]).addClass("active");
                    } else {
                        $(visibleList[selectedIndex + 1]).addClass("active");
                    }
                }
                break;
            }
        }

        if (event.ctrlKey && event.keyCode === 83){ //Ctrl + S
            event.preventDefault();
            if (app.get("currentFileName") === ""){
                $("#save").click();
            } else {
                $.ajax({
                    url: rootPath + "file/object/" + app.get("currentFileName"),
                    type: "PUT",
                    data: network,
                }).done(function(){
                    app.set("execstat", "Saved");
                    updateFileList();
                }).fail(function(data){
                    console.log("failed", data);
                    app.set("execstat", data.responseText);
                });
            }
        }
        if (event.ctrlKey && event.keyCode === 86){ //Ctrl + V
            $(".node").removeClass("nodeSelected");
            _.each(clipboard.nodes, function(add){
                var oldid = add.id;

                var newnode = JSON.parse(JSON.stringify(add));
                var nodecount = _.filter(network.nodes, function(node){
                    // console.log(node);
                    return node.name == newnode.name;
                }).length + 1;

                nodecount = _.find(_.range(nodecount), function(c){
                    if (_.find(network.nodes, function(n){
                        console.log(n.id, "node_" + newnode.name + "_" + (c + 1));
                        return n.id === "node_" + newnode.name + "_" + (c + 1);
                     }) === undefined){
                        return true;
                    } else {
                        return false;
                    }
                }) + 1;
                newnode.id = "node_" + newnode.name + "_" + nodecount;
                newnode.sheet = currentTab;

                clipboard.links = _.map(clipboard.links, function(l){
                    if (l.fromNodeId === oldid){
                        l.fromNodeId = newnode.id;
                    }
                    if (l.toNodeId === oldid){
                        l.toNodeId = newnode.id;
                    }
                    return l;
                });

                network.nodes.push(newnode);
                addNodeDOM(newnode);
                $("#" + newnode.id).addClass("nodeSelected");
            });

            _.each(clipboard.links, function(l){network.links.push(l);});
        }

        if (event.ctrlKey && event.keyCode === 65 && $("#nodelistSearch input:focus").length === 0){ //Ctrl + A
            $(".nodeSelected").removeClass("nodeSelected");
            $("#nodelistSearch input").blur();
            _.chain(network.nodes)
             .filter(function(n){return n.sheet === currentTab;})
             .each(function(n){
                $("#" + n.id).addClass("nodeSelected");
             });
            event.preventDefault();
            event.stopPropagation();
        }
    });
}

var adjustScroll = _.throttle(_adjustScroll, 1000);
function _adjustScroll(sheetname, filter){
    sheetname = sheetname || currentTab;
    filter = filter || function(n){return true;};
    console.log("scrolling");
    var margin = 10;
    var poses = _.chain(network.nodes)
                 .filter(function(n){return n.sheet === sheetname;})
                 .filter(filter)
                 .map(function(n){return _.extend($("#" + n.id).position(),
                                                  {id: n.id});}).value();
    var lefts = _.pluck(_.sortBy(poses, function(p){return p.left;}), "left");
    var tops  = _.pluck(_.sortBy(poses, function(p){return p.top ;}), "top");
    var scroll = {scrollLeft: $("#sheet_" + sheetname).scrollLeft(),
                  scrollTop : $("#sheet_" + sheetname).scrollTop()};

    var obj = $("#sheet_" + sheetname);
    var max = {
        top:  obj.height() + obj.offset().top,
        left: obj.width()  + obj.offset().left
    };

    // top
    // if (_.chain(lefts).map(function(p){return p - margin < 0;}).all().value()){
    //     scroll.scrollLeft += _.min(lefts) - margin;
    // }
    // else if (_.chain(lefts).map(function(p){return p + margin > max.left;}).all().value()){
    //     scroll.scrollLeft += _.max(lefts) - margin;
    // }
    // if (_.chain(tops).map(function(p){return p - margin < 0;}).all().value()){
    //     scroll.scrollTop += _.min(tops) - margin;
    // }
    // else if (_.chain(tops ).map(function(p){return p + margin > max.top ;}).all().value()){
    //     scroll.scrollTop += _.max(tops) - margin;
    // }
    if (!_.sortBy(poses, function(p){return p.left;})[0]) {return;}
    leftest = _.sortBy(poses, function(p){return p.left;})[0].id;
    toppest = _.sortBy(poses, function(p){return p.top; })[0].id;
    // console.log(leftest, toppest);
    scroll.scrollLeft += - ($("#sheets").width()  - $("#" + leftest.id).width() ) / 2 +
                           $("#" + leftest).position().left;
    scroll.scrollTop  += - ($("#sheets").height() - $("#" + toppest.id).height()) / 2 +
                           $("#" + toppest).position().top;

    scroll.scrollLeft = (scroll.scrollLeft > 0) ? scroll.scrollLeft : 0;
    scroll.scrollTop = (scroll.scrollTop > 0) ? scroll.scrollTop : 0;

    // console.log(scroll);
    $("#sheet_" + sheetname).animate(scroll, {easing: "easeOutCubic"});
}

function evalPx(a, b, operator, postfix){
    a = parseInt(a, 10);
    b = parseInt(b, 10);
    switch (operator){
        case '+': return (a + b) + postfix;
        case '-': return (a - b) + postfix;
        case '*': return (a * b) + postfix;
        case '/': return (a / b) + postfix;
        default:
        // console.log("undefined operator");
        return NaN;
    }
}

function addSheetScrollEvent(sheetname){
    scrolls[sheetname] = {top:  $("#sheet_" + sheetname).scrollTop(),
                          left: $("#sheet_" + sheetname).scrollLeft()};
    $("#sheet_" + sheetname).scroll(function(event){
        var tmpscroll = {top:  $("#sheet_" + sheetname).scrollTop(),
                         left: $("#sheet_" + sheetname).scrollLeft()};
        var diff = {top:  tmpscroll.top  - scrolls[sheetname].top,
                    left: tmpscroll.left - scrolls[sheetname].left};
        _.chain(network.nodes)
         .filter(function(n){return n.sheet === sheetname;})
         .each(function(n){
            n.pos[0] -= diff.left;
            n.pos[1] -= diff.top;

            // $("#" + n.id).css("left", evalPx($("#" + n.id).css("left"), diff.left, "-", "px"));
            // $("#" + n.id).css("top", evalPx($("#" + n.id).css("top"), diff.top, "-", "px"));
            $("#" + n.id).offset({left: n.pos[0], top: n.pos[1]});

            var sheetpos = {top: $("#sheet_" + sheetname).offset().top,
                            left: $("#sheet_" + sheetname).offset().left};
            sheetpos.bottom = sheetpos.top  + $("#sheet_" + sheetname).height();
            sheetpos.right  = sheetpos.left + $("#sheet_" + sheetname).width();

            // if (n.pos[1] < $("#sheet_" + sheetname).offset().top){
            //     $("#" + n.id).css("display", "none");
            // } else {
            //     $("#" + n.id).css("display", "block");
            // }
        });
        scrolls[sheetname] = tmpscroll;
    });
}

function removeSheetScrollEvent(sheetname){
    delete scrolls[sheetname];
    $("#sheet_" + sheetname).unbind("scroll");
}

function setupIntroJS(){
    var counter = 1;
    function helper(selector, description){
        $(selector).attr("data-intro", description);
        $(selector).attr("data-step", counter);
        counter += 1;
    }
    $("*").removeAttr("data-intro");
    $("*").removeAttr("data-step");
    if (userPreference.get("lang") === "en"){
        helper("#sheets", "You can build your network here.");
        helper("#addSheetButton", "You can add a new sheet from here");
        helper("#reshapeButton", "You can reshape the network (Graphiz required)");
        helper("#sidebar", "Here is the list of available nodes");
        helper("#nodelistSearch input", "You can search a node from here.");
        helper("#saveButton", "You can save current network here.");
        helper("#executeButton", "You can also run your network");
        helper("#console", "If you have some error, the message will be displayed here");
        helper("#fileManagerButton", "You can upload/download all of your files from here");
        helper("#preferenceButton", "You can configure, e.g., look&feel, language");
        helper("#zoomSlider", "You can zoom up or zoom out with this slider");
    } else if (userPreference.get("lang") === "ja"){
        helper("#sheets", "この画面でネットワークを作成できます");
        helper("#addSheetButton", "ここから新たなシートを追加できます");
        helper("#reshapeButton", "このボタンでノードの整列ができます (Graphvizが必要)");
        helper("#sidebar", "ここに使用できるノードのリストがあります");
        helper("#nodelistSearch input", "ここでノードの検索ができます");
        helper("#saveButton", "現在のネットワークを保存できます。");
        helper("#executeButton", "現在のネットワークを実行できます");
        helper("#console", "エラーがあればここに表示されます。このコンソールは移動できます");
        helper("#fileManagerButton", "ここからすべてのファイルのアップロードやダウンロードができます");
        helper("#preferenceButton", "見た目や言語などの設定ができます");
        helper("#zoomSlider", "このスライダでノードの拡大・縮小ができます");
    } else {
        alert("unknwon language" + userPreference.get("lang"));
    }
}

function forScreenshot(){
    $("#topmenu,#sidebar,#console,#nodelistSearch,#footer").css("display", "none");
    $("#container").css("padding", "0px").css("margin", "0px");
    $("#sheets").css("padding", "0px").css("margin", "0px");
    $("body").css("padding", "0px").css("margin", "0px");
    $("html,body,#container,#sheets,.sheetContainer").css("height", "90%");
    $("#sheets").tabs("refresh");
}

function subnetParamMaintain(nodeid){
    var nodeobj = _.findObject(network.nodes, "id", nodeid);
    // console.log(nodeobj.parameters);
    if (nodeobj.sheet === "MAIN"){
        return;
    }
    _.each(nodeobj.parameters, function(p){
        if (p.type !== "subnet_param"){return;}
        if (p.value === ""){return;}

        var addparam = function(n){
            var origType = _.findObject(
                _.findObject(nodedefs, "name", nodeobj.name).parameters,
                "name", p.name);
            origType = (origType) ? origType.type : "subnet_param";
            if (n.name === nodeobj.sheet &&
               !_.findObject(n.parameters, "name", p.value)){
                n.parameters.push({
                    condition: "",
                    description: "subnet_param",
                    list: [],
                    name: p.value,
                    type: origType,
                    value: ""            // same as ...
                });
            }
        };
        _.each(network.nodes, addparam);    // add to nodes
        _.each(nodedefs, addparam);         // add to nodedef
    });
    removeNonExistingSubnetParam();
}

function removeNonExistingSubnetParam(){
    // remove all non-existing subnet param from dynamic
    var filterNonExistingSubnetParam = function(nodeobj){
        return _.filter(nodeobj.parameters, function(paramobj){
            return _.chain(network.nodes)
                    .filter(function(n){return n.sheet === nodeobj.name;})
                    .pluck("parameters")
                    .flatten()
                    .filter(function(p){return p;})
                    .filter(function(p){return p.type === "subnet_param";})
                    .find(function(p){return paramobj.name === p.value;}).value() !== undefined;
            });
    };

    _.each(network.nodes, function(n){
        if (n.category === "Dynamic"){
            n.parameters = filterNonExistingSubnetParam(
                JSON.parse(JSON.stringify(n)));
        }
    });
    _.each(nodedefs, function(n){
        if (n.category === "Dynamic"){
            n.parameters = filterNonExistingSubnetParam(
                JSON.parse(JSON.stringify(n)));
        }
    });
}
